/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/* 
 * $Id: conditional.cpp,v 1.14 2002/06/15 15:16:13 mulix Exp $
 */

#include "conditional.h"
#include "tokens_to_types.h"
#include "logger.h"
#include "validity_checks.h"
#include "parser.h"
#include "parse_exception.h"

#include <cstring>


conditional::conditional()
{
    _op = OT_LAST_OPERATOR;
    _val.type = DT_LAST_TYPE;
    _val.val.i = 0;
}

conditional::~conditional()
{
    if (_val.type == DT_STRING) 
	delete[] (_val.val.str);
}

conditional::conditional(const conditional& rhs)
{
    /* TODO: handle this with a suitable exception */
    if (!copy_value_type(_val, rhs._val))
	throw string("bad things happend, couldn't copy value type"); 

    _op = rhs._op;
    _tmp_val = rhs._tmp_val; 
    _line_num = rhs._line_num; 
    _data_line_num = rhs._data_line_num; 
}

ostream& 
conditional::print(ostream& os)
{
    return os << tokens::get_operator_str(_op) << " " << _val << endl;
}


bool 
conditional::validate(const string& sys_name)
{
    if (_op == OT_LAST_OPERATOR) {
	strstream err; 
	err << "no operator supplied for syscall '" << sys_name << "'" << ends;
	throw validate_exception("", _line_num, err.str());
    } 
    if (_val.type == DT_LAST_TYPE) {
	strstream err; 
	err << "no data type found for param at syscall '" << sys_name << "'" << ends; 
	throw validate_exception("", _line_num, err.str());
    }
    
    /* TODO: make it into a void function? */
    return true;
}

void 
conditional::set_operator(const string& str)
{
    _op = tokens::get_operator(str);
    if (_op == OT_LAST_OPERATOR) {
	strstream err; 
	err << "'" << str << "' is not an operator" << ends;
	throw parse_exception("", get_line_num(), err.str());
    }
}

void 
conditional::set_operator(oper_type_t op)
{
    _op = op;
}

oper_type_t
conditional::get_operator() const
{
    return _op;
}

static bool 
is_integer(sct_data_type_t dt)
{
    switch (dt) {
    case(DT_CHAR):
    case(DT_INT):
    case(DT_UINT):
    case(DT_SHORT):
    case(DT_USHORT):
    case(DT_LONG):
    case(DT_ULONG): 
	/* we fall through to here from the top */
	return true;
    case(DT_FLOAT):
    case(DT_STRING):
    case(DT_LAST_TYPE):
    default:
	/* we fall through to here from DT_FLOAT */
	return false; 
    }
    
}

void 
conditional::set_value(const string& str)
{
    /* if we get a quoted string, strip the quotes */ 
    if (str[0] == '"' && str.length() > 1 && str[str.length()-1] == '"')
	_tmp_val = str.substr(1, str.length() - 2);
    else
	_tmp_val = str;
    
}

const value_type_t&
conditional::get_value() const
{
    return _val;
}

sct_filter_node_t* 
conditional::sct_get_filter_node(sct_filter_node_t* lhs) const
{
    sct_filter_node_t* res = NULL; 

    assert(lhs);

    if (!lhs)
	return res;
    
    sct_filter_node_t* rhs = value_type_t_to_filter_node(get_value());
    if (!rhs)
	return res; 

    res = operator_to_filter_node(_op, lhs, rhs);
    
    return res;
}

bool
conditional::finalize(const string& sys_name)
{
    /* deduce the data type for our value */
    sct_data_type_t required_data_type = query_required_data_type(sys_name);
    if (required_data_type == DT_LAST_TYPE)
	goto value_error;
    
    if (is_integer(required_data_type) &&!is_int(_tmp_val))
	goto value_error;
    
    switch (required_data_type){
    case(DT_BOOL):
	if (!strcasecmp(_tmp_val.c_str(), "true")) {
	    _val.val.b = true;
	} else if (!strcasecmp(_tmp_val.c_str(), "false")) {
	    _val.val.b = false;
	} else {
	    goto value_error;
	}
        break;
    case(DT_CHAR):
	_val.val.c = (char)atoi(_tmp_val.c_str());
	break;
    case(DT_INT):
	_val.val.i = atoi(_tmp_val.c_str());
	break;
    case(DT_UINT):
	_val.val.ui = (unsigned int)atoi(_tmp_val.c_str());
	break;
    case(DT_SHORT):
	_val.val.s = (short)atoi(_tmp_val.c_str());
	break;
    case(DT_USHORT):
	_val.val.us = (unsigned short)atoi(_tmp_val.c_str());
	break;
    case(DT_LONG):
	_val.val.l = atol(_tmp_val.c_str());
	break;
    case(DT_ULONG):
	_val.val.ul = (unsigned long)atol(_tmp_val.c_str());
	break;
    case(DT_FLOAT):
	if (!is_float(_tmp_val))
	    goto value_error;

	_val.val.f = atof(_tmp_val.c_str());
	break;
    case(DT_STRING):
	/* if allocation fails, new throws an exception */
	_val.val.str = new char[_tmp_val.size()+1]; 
	strcpy(_val.val.str, _tmp_val.c_str());
	break;
    case(DT_LAST_TYPE):
    default:
	goto value_error;
    }

    /* function ends here, unless we had an error */
    _val.type = required_data_type;
    return true; 

 value_error:
    strstream buf;
    buf << "conditional::finalize called with invalid value parameter.\n"
	<< "this conditional has param type: '" 
	<< tokens::get_data_type_str(required_data_type) 
	<< "', supplied value is: '" <<  _tmp_val << "'" << ends; 
    
    throw validate_exception("",_line_num, buf.str());
}
