/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 * $Id: filter_parser.h,v 1.17 2002/06/17 23:30:33 mulix Exp $
 */

#ifndef FILTER_PARSER_H_
#define FILTER_PARSER_H_

#include <vector>
#include <map>
#include <string> 

#include "sct_rules/rule_engine_public.h"
#include "parser.h"
#include "parse_exception.h"
#include "param_types.h"

using namespace std;

class filter_parser{
public:
    /* a vector of strings */
    typedef vector < string > str_tok_vec;

    /* the types of tokens this parser knows */
    enum token_t {
	NUMBER = 0, STRING, LPAREN, RPAREN,
	LOGICAL_NOT, BITWISE_NOT, 
	PLUS, MINUS, 
	SHIFT_LEFT, SHIFT_RIGHT, 
	LESS_THEN, LESS_OR_EQUAL, GREATER_THEN, GREATER_OR_EQUAL,
	EQUAL, NOT_EQUAL, PATTERN_MATCH, 
	BITWISE_AND, BITWISE_XOR, BITWISE_OR, 
	LOGICAL_AND, LOGICAL_OR, IN_OP,
	UID, PID, EUID, SUID, GID, EGID, SGID, COMM, 
	PPID, PCOMM, RETVAL, 
	LBRACKET, RBRACKET, PARAMS, COMMA,
	DOT,
	END, BUILTIN_MACRO, BUILTIN_PARAMLESS_MACRO, LAST_TOKEN_TYPE
    };

    struct fl_token {
	fl_token();
	~fl_token();
	fl_token(const fl_token& rhs);
	fl_token& operator=(const fl_token& rhs);
	string orig_text;
	token_t t;
	value_type_t v; 
    };

    /* a macro handler */
    typedef sct_filter_node_t*
	(filter_parser::* macro_handler_func_t)(const fl_token& p1);

    typedef struct {
	macro_handler_func_t macro;
    } macro_handler_t;

    typedef map < string, macro_handler_t* > macro_handlers_map;


    filter_parser(const string& syscall_name);

    int get_line_num() {return _line_num;}
    void set_line_num(int line_num) {_line_num = line_num;}

    /* parse this string into a sct_filter_node_t*. */
    void parse(sct_filter_node_t*& f, const string& data);
private:
    /* macro function registration. */
    void filter_parser::register_macro(const char* macro_name,
				       macro_handler_func_t func);
    void register_noparam_macro(const char* macro_name,
				  macro_handler_func_t func);

    bool is_operator(const string& str);

    /* advances the curr_tok_str iterator */
    void get_token(fl_token& tok);

    /* peek at the next in line token, without advancing the iterator */
    void peek_token(fl_token& tok);

    /* pop the next token, disregarding its value */
    void pop_token();

    /* push the current token back (decremment the curr_tok_str iterator) */
    void push_token();

    /* build a token from the current position of the curr_tok_str iterator */
    void build_token(str_tok_vec::const_iterator& pos, fl_token& tok);

    sct_filter_node_t* prim(bool get);
    sct_filter_node_t* term(bool get);
    sct_filter_node_t* expr(bool get);
    sct_filter_node_t* shift(bool get);
    sct_filter_node_t* order(bool get);
    sct_filter_node_t* equality(bool get);
    sct_filter_node_t* expr_bitwise_and(bool get);
    sct_filter_node_t* expr_bitwise_xor(bool get);
    sct_filter_node_t* expr_bitwise_or(bool get);
    sct_filter_node_t* expr_logical_and(bool get);
    sct_filter_node_t* expr_logical_or(bool get);
    sct_filter_node_t* vector(bool get);

    sct_filter_node_t* parse_param();
    sct_data_type_t get_data_type_for_param(unsigned int param_index);

    sct_filter_node_t* parse_builtin_macro(const string& macro_name,
					   bool has_param);
    sct_filter_node_t* calc_ipaddr(const fl_token& p1);
    sct_filter_node_t* calc_htons(const fl_token& p1);
    sct_filter_node_t* calc_usernametoid(const fl_token& p1);
    sct_filter_node_t* calc_groupnametoid(const fl_token& p1);

    sct_filter_node_t* calc_macro_open_flags(const fl_token& p1);
    
    map< string, token_t > strings_to_tokens;

    macro_handlers_map macro_handlers;

    str_tok_vec str_toks; 
    str_tok_vec::const_iterator curr_tok_str; 

    int number_value;
    string str_value;
    bool started; 
    string _syscall_name;
    int _line_num;
    int _data_line_num;

    fl_token curr_tok;
};

#endif /* FILTER_PARSER_H_ */
