/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 * $Id: main.cpp,v 1.36 2002/06/15 15:16:13 mulix Exp $
 */

/* This is the main entry point into the system call tracker 
 * configuration utility
 * @author: mulix@actcom.co.il
 */

#include <iostream>
#include <fstream>
#include <strstream>
#include <list>
#include <cstdio>

extern"C"{
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h> //for stat()
#include <sys/types.h> //for uid_t
#include <getopt.h> //for getopt 
}

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/rule_engine_public.h"
#include "rule.h"
#include "logger.h"
#include "parser.h"
#include "syscalls.h"
#include "tokens_to_types.h"

const char DEFAULT_CONFIG_FILE_NAME[] = "sct_example.conf";
const char DEFAULT_SYSCALLS_INFO_LOCATION[] = "../sct_rules_module/syscalls.dat";

const char* progname; 

/* forward declarations */
bool load_config_file(const char * config_file_name,
		      strstream& data);
bool parse_config_file(strstream& data,
		       list<rule>& rules, 
		       const string& config_file_name);

#ifdef SCT_CONFIG_DEBUG
int debug = 1;
#else /* SCT_CONFIG_DEBUG is not defined */
int debug = 0;
#endif

void usage(const char* name)
{
    cerr << name << ": a configuration and control utility for the " 
	 "syscall tracker" << endl;

    cerr << "Usage:\t" << name << " [-v|--verbose] "
	"[-c|--control device_file|sysctl] command [file name]" << endl;
    
    cerr << "where 'command' is one of: " << endl;
    cerr << "print: print all currently registered rules" << endl;
    cerr << "delete: delete all currently registered rules" << endl;
    cerr << "upload [file name]: parse configuration file "
	 << "and upload rules to the kernel module" << endl;
    cerr << "check [file name]: parse and print all rules "
	 << "in the configuration file, but don't upload "
	 << "them to the kernel module" << endl;
    cerr << "download: get and print all of the currently "
	 << "registered rules" << endl; 
    cerr << "count: get the number of currently registered "
	    "rules" <<endl; 
    exit(1);
}

static int 
string_to_impl(const char* str, sct_ctrl_impl_t* impl)
{
    if (!strcmp(optarg,"device_file"))
	*impl = SCT_CTRL_DEVICE_FILE; 
    else if (!strcmp(optarg,"sysctl"))
        *impl = SCT_CTRL_SYSCTL; 
    else
	return 0; 
    
    return 1; 
}

void parse_options(char** argv, int argc, bool& print, 
		   bool& del, bool& check, bool& upload, 
		   bool& verbose, bool& download, 
		   bool& count, const char*& fname, 
		   sct_ctrl_impl_t* impl)
{
    int option_index = 0;
    char* cmd; 

    print = del = check = upload = verbose = download = count = false;

    while(1){ 
	/* structure with all recognised options for sct_config */
	static struct option long_options[] = {
	    {"verbose", 0, 0, 0},  
	    {"control", 1, 0, 0},  
	    {0, 0, 0, 0}
	};
	int c;
	c = getopt_long (argc, argv, "vc:", long_options, &option_index);

	switch (c) {
	case 0: /* long option specified */ 
	    switch (option_index) {
	    case 0: /* verbose */
		verbose = true; 
		break;
	    case 1:
		if (!string_to_impl(optarg, impl))
		    usage(progname); 
		break; 
	    }
	    break;
	case 'v':
	    verbose = 1; 
	    break; 
	case 'c':
	    if (!string_to_impl(optarg, impl))
		usage(progname); 
	    break; 
	case '?': /* unrecognised option */
            /* fall through */
        default:
            c = -1;
            break;
	}
	if(c == -1) 
	    break;  /* no more options */
    }

    argc -= optind; 
    if (argc <= 0) { /* no command specified */
	usage(progname); 
    }

    argv += optind; 
    cmd = argv[0]; 
    
    if (!strcasecmp(cmd, "print")){
	print = true; 
    } else if (!strcasecmp(cmd, "delete")){
	del = true; 
    } else if (!strcasecmp(cmd, "upload")){
	upload = true;
    } else if (!strcasecmp(cmd, "check")){
	check = true;
    } else if (!strcasecmp(cmd, "download")) {
	download = true; 
    } else if (!strcasecmp(cmd, "count")) {
	count = true; 
    } else 
	usage(progname);

    if (upload || check) {
	fname = (--argc > 0) ? argv[1] : DEFAULT_CONFIG_FILE_NAME;
    }
}

void 
print_rule_list(syscall_rules_t* rules)
{
    assert(rules); 

    for ( ; rules; rules = rules->next) {
	char sid[32]; 
	sct_get_syscall_id_as_str(rules->syscall_id, sid, sizeof(sid)); 
	cout << "syscall: " << sid 
	     << ", " << (rules->when == RW_BEFORE? "before" : "after") 
	     << " rules:" << endl; 
	sct_print_tracking_rules(rules->list); 
    }
}

void 
free_rule_list(syscall_rules_t* rules)

{
    int first = 1; /* we dont free the head of the list */ 

    assert(rules); 

    for ( ; rules; rules = rules->next) {
	sct_delete_tracking_rule_list(rules->list); 
	rules->list = NULL;
	if (!first)
	    free(rules); 
	first = 0; 
    }
}

static int
get_module_rules(sct_ctrl_t ctrl)
{
    int rc; 
    syscall_rules_t rules; 

    rc = sct_ctrl_get_all_rules(ctrl, &rules); 
    if (rc == -1) {
	logger::log(logger::LOG_LEVEL_ERROR, 
		    "Error: failed getting rules.\n"); 
	return 0; 
    }
    if (rc == 0) { /* empty list */ 
        fprintf(stdout, "There are no module rules defined\n"); 
	return 1; 
    }
    /* goodie! we have the list of rules! */ 
    /* now let's print them */ 
    print_rule_list(&rules); 
    free_rule_list(&rules); 

    return 1; 
}

static int 
print_module_rules(sct_ctrl_t ctrl)
{
    int rc = sct_ctrl_print_all_rules(ctrl);
    if (!rc){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "Error: failed printing all rules.\n");
    }

    return rc; 
}

static int 
delete_all_rules(sct_ctrl_t ctrl)
{
    int rc = sct_ctrl_del_all_rules(ctrl);
    if (!rc){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "Error: failed deleting all previous module rules.\n");
    }

    return rc;
}

static int
get_module_rules_count(sct_ctrl_t ctrl)
{
   int rc = sct_ctrl_get_all_rule_count(ctrl); 
   if (rc == -1) {
	   logger::log(logger::LOG_LEVEL_ERROR, 
		       "Error: failed getting the rule count.\n"); 
	   return 0; 
   }
   printf("The module currently has %d rules registered\n", rc); 
   return 1; 
}

int main(int argc, char ** argv)
{
    bool print, erase, check, upload, verbose, download, count;
    const char* config_file_name; 
    sct_ctrl_impl_t impl = SCT_CTRL_DEFAULT; 

    int rc;

    progname = argv[0];

    if (argc < 2)
	usage(progname);

    parse_options(argv, argc, print, erase, check,
		  upload, verbose, download, count, 
		  config_file_name, &impl);

    if (verbose)
	logger::set_verbosity_level(logger::LOG_LEVEL_INFO);

    /* init the control library */
    sct_ctrl_t sct_ctrl;
    if (!check){
        sct_ctrl = sct_ctrl_init(impl);
	if (!sct_ctrl){
	    logger::log(logger::LOG_LEVEL_ERROR,
			"failed to initialize the ctrl library\n");
	    return 1;
	}
    }
    if (count) {
        rc = get_module_rules_count(sct_ctrl); 
	if (!rc)
		cout << "getting the count of module rules failed\n" << endl; 
	return (!rc); 
    }
     
    if (print) {
	rc = print_module_rules(sct_ctrl);
	if (rc)
	    cout << "module rules printed to system log file" << endl;
	return (!rc);
    }

    if (download) {
	rc = get_module_rules(sct_ctrl); 
	if (!rc)
	    cout << "getting module rules failed" <<endl; 
	return (!rc); 
    }

    if (erase) {
	rc = delete_all_rules(sct_ctrl); 
	if (rc)
	    cout << "all module rules were deleted sucessfully" << endl;
	return (!rc);
    }

    /* it's either check or upload. let's go on */

    strstream data;
    if (!strcmp(config_file_name, "-")) {
	cin>>data.rdbuf();
    } else if (!load_config_file(config_file_name, data)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "failed to read the config file '%s'. quitting.\n", 
		    config_file_name);
	return 1;
    }
    //the list of rules
    list<rule> rules;

    if (!parse_config_file(data,
			   rules,
			   config_file_name)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "failed to parse the config file. quitting.\n");
	return 1;
    }

    if (check || upload) {
	cout << "parsed rule list:" << endl;
	list<rule>::const_iterator it = rules.begin();
	for (; it != rules.end(); ++it){
	    cout << *it << endl;
	}
	if (check)
	    return 0;
    }

    rc = delete_all_rules(sct_ctrl);
    if (!rc) {
	return 1;
    }

    list<rule>::const_iterator it = rules.begin();
    for (; it != rules.end(); ++it){
	strstream rule_data;
	rule_data << *it << '\0';
	logger::log(logger::LOG_LEVEL_INFO,
		    "handling rule:\n%s",
		    rule_data.str());

	sct_filter_node_t* filter = (*it).filter_node();

	sct_action_t* act =  (*it).get_action();

	rc = sct_ctrl_add_rule(sct_ctrl, (*it).get_rule_id(), 
			       (*it).get_rule_when(),
			       (*it).get_syscall_identifier(), 
			       filter, act, 1);
	if (!rc){
	    logger::log(logger::LOG_LEVEL_ERROR,
			"Error: failed adding rule '%s'\n",
			(*it).get_rule_name().c_str());
	    return -1;
	}
	sct_free_action(act);
    }
    cout << "successfully uploaded rules from file '"
	 << config_file_name << "'" << endl;
     
    return 0;
}


bool load_config_file(const char * config_file_name,
		      strstream& data)
{
    //first of all, make some sanity checks on the file

    //preferably, that user will be root, but we relax that requirement for
    //development
    struct stat stat_info;
    if (stat(config_file_name, &stat_info)==-1){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "load_config_file: stat '%s' failed, with error '%s'\n",
		    config_file_name, strerror(errno));
	return false;
    }

    //the config file is a regular file
    if (!S_ISREG(stat_info.st_mode)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "load_config_file: file '%s' is not a regular file\n", 
		    config_file_name);
	return false;
    }
#if 0
    //the config file is owned by the current user
    if (stat_info.st_uid != geteuid()){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "load_config_file: file is not owned by the current euid\n");
	return false;
    }

    //the config file is only writeable by the current user. 
    if ((stat_info.st_mode & (S_IRUSR | S_IWUSR)) != (S_IRUSR & S_IWUSR)){
	log("load_config_file: wrong permissions on config_file");
	log(config_file_name);
	return false;
    }
#endif

    //now, that we've verified the file is usable by us, let's load
    //its contents into a buffer.
    ifstream in(config_file_name);
    in>>data.rdbuf();
    return true;
}

bool 
parse_config_file(strstream& data,
		  list<rule>& rules,
		  const string& config_file_name)
{
    /* maps from string to enumerated types     */
    /* must be done here since the syscalls.dat */
    /* parser uses them as well                 */
    tokens::init();
    
    // load the syscall information file
    if (!syscalls::load(DEFAULT_SYSCALLS_INFO_LOCATION)) {
	logger::log(logger::LOG_LEVEL_ERROR,
		    "Error: failed to find syscall information "
		    "file at '%s'.\n", DEFAULT_SYSCALLS_INFO_LOCATION);
	return false;
    }
    
    //create a new parser
    parser p(data, config_file_name);
    //start parsing the config data
    return p.parse(rules);
}
