/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 *$Id: param_conditional.cpp,v 1.28 2002/06/15 15:16:13 mulix Exp $
 */

#include <string>
#include <stdlib.h>

#include <vector>
#include <utility> 
#include <map>

#include "parser.h"
#include "parse_exception.h"
#include "param_conditional.h"
#include "logger.h"
#include "token.h"
#include "validity_checks.h"
#include "syscalls.h"
#include "tokens_to_types.h"

using namespace std;

param_conditional::param_conditional()
{
    _loc = 0; /* bogus */
    _loc_field = "";
    _loc_field_index = -1;
}

param_conditional::~param_conditional()
{
}

param_conditional::param_conditional(const param_conditional& rhs):
    conditional(rhs)
{
    _loc = rhs._loc;
    _loc_field = rhs._loc_field;
    _loc_field_index = rhs._loc_field_index;
}


ostream& 
param_conditional::print(ostream& os)
{
    os<<"param conditional: "<<endl;
    os<<"\t\tparam: "<<_loc<<"." <<_loc_field<<" ";
    return conditional::print(os)<<endl;
}

void 
param_conditional::set_location(const string& str)
{
    if (!is_int(str)) {
	string err = "'" + str + "' is not a number";
	throw parse_exception("", get_line_num(), err);
    }
    _loc = atoi(str.c_str());
    if (_loc == 0) {
	string err = "'" + str + "' is an invalid location";
	throw parse_exception("", get_line_num(), err);
    }
}

void param_conditional::set_location_field(const string& str)
{
    _loc_field = str;
}

void param_conditional::set_location_field_index(int field_index)
{
    _loc_field_index = field_index;
}

bool
param_conditional::is_valid_for_finalization(const string& sys_name)
{
    /* is there such a syscall? */
    if (!syscalls::has_syscall(sys_name)) {
	string err = "no such syscalls '" + sys_name + "'";
	throw validate_exception("", get_line_num(), err);
    }

    const syscalls::param* p = syscalls::syscall_param_at(sys_name, _loc);
    unsigned int num_params = syscalls::syscall_params(sys_name).size();

    /* no such parameter? */
    if (!p) {
	strstream buf; 
	buf << "location given is '" << _loc << "', syscall '" << sys_name
	    <<"' only has '" << num_params << "' params" << '\0';
	throw validate_exception("", get_line_num(), buf.str());
    }
    
    /* check that it's not an unmatchable param, 
       and that the datatype is the same 
       (add casting later?) */
    if (p->name == "INVALID"){
	strstream buf;
	buf << "match requested on unmatchable parameter for syscall '"
	    << sys_name << "' at parameter location '" 
	    << _loc << "'" << '\0';
	throw validate_exception("", get_line_num(), buf.str());
    }
    
    // handle structs
    if (p->is_struct) {
	const string& struct_type = p->struct_type;
        /* is there such a struct type? */
        if (!syscalls::has_struct(struct_type)) {
	    strstream buf; 
	    buf << "no such struct type '" << struct_type << "'\0";
	    throw validate_exception("", get_line_num(), buf.str());
        }
        /* no such parameter? */
	int field_index = 0;
	const syscalls::param* p =
	    syscalls::struct_param_at(struct_type, _loc_field, &field_index);

        if (!p) {
	    strstream buf;
	    buf << "struct field given ' " << _loc_field << "' not found in struct '"
		<< struct_type << "'"<<'\0';
	    throw validate_exception("", get_line_num(), buf.str());

        }
        /* check that it's not an unmatchable struct param, 
           and that the datatype is the same 
           (add casting later?) */
        if (p->name == "INVALID"){
	    strstream buf;
	    buf << "match requested on unmatchable parameter "
		<< "for syscall '" << sys_name << "' at parameter location '"
		<< _loc << "', struct field ' " << _loc_field << "'\0";
	    throw validate_exception("", get_line_num(), buf.str());
	}
    }
    
    return true;
}

bool
param_conditional::validate(const string& sys_name)
{
    return conditional::validate(sys_name);
}

unsigned short int
param_conditional::get_location(void) const
{
    return _loc;
}

int
param_conditional::get_location_field_index(void) const
{
    return _loc_field_index;
}

sct_filter_node_t*
param_conditional::sct_get_filter_node(void) const
{
    /* sct_make_var_indexed_field_val expects the location to be 0 based */
    sct_filter_node_t* rhs = 
	sct_make_var_indexed_field_val(VT_PARAM, _loc - 1, 
				       _loc_field_index, 
				       _val.type, 
				       -1 /* struct_type will be static */);

    if (!rhs)
	return NULL; 
    
    return conditional::sct_get_filter_node(rhs);
}

bool
param_conditional::finalize(const string& sys_name)
{
    if (!is_valid_for_finalization(sys_name)){
	return false; 
    }

    int field_index = -1;  /* assume its a simple param, not a struct. */
    
    const syscalls::param* p = 
	syscalls::get_final_syscall_param(sys_name, _loc, _loc_field, &field_index);
    if (!p)
	return false; 

    set_location_field_index(field_index);

    return conditional::finalize(sys_name);
}

sct_data_type_t 
param_conditional::query_required_data_type(const string& sys_name)
{
    int dummy;
    const syscalls::param* p = 
	syscalls::get_final_syscall_param(sys_name, _loc, _loc_field, &dummy);
    if (p)
	return p->type;

    return DT_LAST_TYPE;
}
