/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 *$Id: param_types.cpp,v 1.20 2002/06/15 15:16:13 mulix Exp $
 */

#include <iostream>
#include <assert.h>

#include "param_types.h"
#include "tokens_to_types.h"
#include "logger.h"

ostream& operator<<(ostream& os, const value_type_t& vt)
{
    const string& dtstr = tokens::get_data_type_str(vt.type);
    os<<"'"<<dtstr<<": ";
    
    switch(vt.type){
    case(DT_BOOL):os<<vt.val.b;
	break;
    case(DT_CHAR):os<<vt.val.c;
	break;
    case(DT_INT):os<<vt.val.i;
	break;
    case(DT_UINT):os<<vt.val.ui;
	break;
    case(DT_SHORT):os<<vt.val.s;
	break;
    case(DT_USHORT):os<<vt.val.us;
	break;
    case(DT_LONG):os<<vt.val.l;
	break;
    case(DT_ULONG):os<<vt.val.ul;
	break;
    case (DT_FLOAT):os<<vt.val.f;
	break;
    case (DT_STRING):os<<(vt.val.str) ? vt.val.str : "NULL";
	break;
    case (DT_LAST_TYPE): /* fallthrough */
    default:
	break;
	/* TODO: what to print here? the data type is undefined
	   so the value has no meaning */ 
    }

    os<<"'";
    return os;
}

ostream& operator<<(ostream& os, const process_field_t& proc)
{
    return os<<tokens::get_process_field_str(proc);
}

value_type_t* clone_value_type(const value_type_t* val)
{
    value_type_t* new_val = NULL;

    if (val) {
	new_val = new value_type_t;
	if (!new_val)
	    goto clone_value_type_return;

	new_val->type = DT_LAST_TYPE;  /* no type yet. */
	if (!copy_value_type(*new_val, *val))
	    goto clone_value_type_free_new_val;

    }

    goto clone_value_type_return;

  clone_value_type_free_new_val:
    if (new_val)
        delete new_val;
    new_val = NULL;

  clone_value_type_return:
    return new_val;
}

bool copy_value_type(value_type_t& val_to, const value_type_t& val_from)
{
    bool success = false;  /* assume failure */

    /* handle strings freeing properly. */
    if (val_to.type == DT_STRING && val_to.val.str) {
	delete[] val_to.val.str;
	val_to.val.str = NULL;
    }

    switch (val_from.type) {
    	    case DT_BOOL:   val_to.val.b = val_from.val.b; break;
    	    case DT_CHAR:   val_to.val.c = val_from.val.c; break;
    	    case DT_INT:    val_to.val.i = val_from.val.i; break;
    	    case DT_UINT:   val_to.val.ui = val_from.val.ui; break;
    	    case DT_SHORT:  val_to.val.s = val_from.val.s; break;
    	    case DT_USHORT: val_to.val.us = val_from.val.us; break;
    	    case DT_LONG:   val_to.val.l = val_from.val.l; break;
    	    case DT_ULONG:  val_to.val.ul = val_from.val.ul; break;
    	    case DT_FLOAT:  val_to.val.f = val_from.val.f; break;
    	    case DT_STRING: {
				val_to.val.str = NULL;
				if (val_from.val.str) {
				    val_to.val.str =
					new char[strlen(val_from.val.str)+1];
				    if (!val_to.val.str) {
					goto copy_value_type_return;
				    }
				    strcpy(val_to.val.str, val_from.val.str);
				}
				break;
			    }
    	    case DT_LAST_TYPE: 
    	    default:
		logger::log(logger::LOG_LEVEL_ERROR,
		    	    "bad data type in clone_value_type - %d\n",
		    	    val_from.type);
		goto copy_value_type_return;
    }

    val_to.type = val_from.type;
    success = true;

  copy_value_type_return:
    return success;
}

void free_value_type(value_type_t* val)
{
    if (val) {
	switch (val->type) {
    	    case DT_STRING:
		if (val->val.str)
		    delete[] val->val.str;
		break;
    	    default:
		/* do nothing. */
		break;
	}
	delete val;
    }
}

sct_filter_node_t* 
value_type_t_to_filter_node(const value_type_t& val)
{
    switch (val.type){
    case DT_BOOL:   return sct_make_bool_val(val.val.b); break;
    case DT_CHAR:   return sct_make_char_val(val.val.c); break;
    case DT_INT:    return sct_make_int_val(val.val.i); break;
    case DT_UINT:   return sct_make_uint_val(val.val.ui); break;
    case DT_SHORT:  return sct_make_short_val(val.val.s); break;
    case DT_USHORT: return sct_make_ushort_val(val.val.us); break;
    case DT_LONG:   return sct_make_long_val(val.val.l); break;
    case DT_ULONG:  return sct_make_ulong_val(val.val.ul); break;
    case DT_FLOAT:  return sct_make_float_val(val.val.f); break;
    case DT_STRING: return sct_make_string_val(val.val.str); break;
    case DT_LAST_TYPE: 
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "bad data type in value_type_t_to_filter_node - %d\n",
		    val.type);
	return NULL;
    }
}

sct_filter_node_t* 
operator_to_filter_node(oper_type_t op, sct_filter_node_t* t1, sct_filter_node_t* t2)
{
    switch(op){
    case OT_EQUAL:           return sct_make_equal(t1, t2);
    case OT_LESS:            return sct_make_less_then(t1, t2);
    case OT_PATTERN_MATCH:   return sct_make_pattern_match(t1, t2);
    case OT_GREATER:         return sct_make_greater_then(t1, t2); 
    case OT_LESS_EQUAL:      return sct_make_less_or_equal(t1, t2);
    case OT_GREATER_EQUAL:   return sct_make_greater_or_equal(t1, t2); 
    case OT_NOT_EQUAL:       return sct_make_not_equal(t1, t2); 
    case OT_PLUS:            return sct_make_math_plus(t1, t2);
    case OT_MINUS:           return sct_make_math_minus(t1, t2);
    case OT_BIT_SHIFT_LEFT:  return sct_make_bit_left_shift(t1, t2);
    case OT_BIT_SHIFT_RIGHT: return sct_make_bit_right_shift(t1, t2);
    case OT_BIT_AND:         return sct_make_bit_and(t1, t2);  
    case OT_BIT_OR:          return sct_make_bit_or(t1, t2);
    case OT_BIT_XOR:         return sct_make_bit_xor(t1, t2);
    case OT_BIT_NOT:         return sct_make_bit_not(t1);
    case OT_LOGIC_AND:       return sct_make_logical_and(t1, t2);
    case OT_LOGIC_OR:        return sct_make_logical_or(t1, t2);
    case OT_LOGIC_NOT:       return sct_make_logical_not(t1);
    case OT_LAST_OPERATOR:     
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "unimplemented or unknown op '%d' in operator_to_filter_node\n",
		    (int)op);
	return NULL;
    }
}

sct_filter_node_t* 
process_field_t_to_var(process_field_t proc)
{
    /* TODO: more matches! */
    switch(proc){
    case PT_UID:  return sct_make_var_val(VT_PROC_UID, DT_USHORT); /* TODO: get the actual dt size */
    case PT_GID:  return sct_make_var_val(VT_PROC_GID, DT_USHORT); 
    case PT_COMM: return sct_make_var_val(VT_PROC_COMM, DT_STRING);
    case PT_PCOMM: return sct_make_var_val(VT_PROC_PCOMM, DT_STRING);
    case PT_PID:  return sct_make_var_val(VT_PROC_PID, DT_LONG);
    case PT_PPID: return sct_make_var_val(VT_PROC_PPID, DT_LONG);
    case PT_EUID: return sct_make_var_val(VT_PROC_EUID, DT_USHORT);
    case PT_SUID: return sct_make_var_val(VT_PROC_SUID, DT_USHORT);
    case PT_EGID: return sct_make_var_val(VT_PROC_EGID, DT_USHORT);
    case PT_SGID: return sct_make_var_val(VT_PROC_SGID, DT_USHORT);
    case PT_LAST_PROCESS_FIELD: /* fallthrough */
    default:
	logger::log(logger::LOG_LEVEL_ERROR,
		    "bad or unimplemented process conditional "
	            "in process_field_t_to_var: %d\n",
		    (int)proc);
	return NULL;
    }
}
