/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/* 
 * $Id: parse_exception.h,v 1.4 2002/06/15 15:16:13 mulix Exp $
 */

#ifndef _PARSE_EXCEPTION_H_
#define _PARSE_EXCEPTION_H_

#include <string>
#include <iostream>

#include "token.h"

using namespace std;

class exception_intf
{
 public:
    virtual void print(ostream& out) const = 0;

    virtual void set_file_name(const string& file_name) = 0;
    virtual const string& get_file_name(void) const = 0;
};

class basic_exception : public exception_intf
{
 public:
    /* constructor */
    basic_exception(const string& conf_file, unsigned int conf_line, 
		    const string& err);

    virtual ~basic_exception();
    
    /* print this excpetion to ostream 'out' */
    virtual void print(ostream& out) const;

    /* get/set the file name */ 
    virtual void set_file_name(const string& file_name) {_conf_fname = file_name;}
    virtual const string& get_file_name(void) const {return _conf_fname; }
    
    
 private:
    /* the configuration file where this error occured */
    string _conf_fname;

    /* the line in the configuration file where this error occured */
    unsigned int _conf_line;

    /* the error message the thrower gave us */
    string _err;
};

class debug_exception: public exception_intf
{
 public:
    debug_exception(const string& fname = "", const string& func = "", 
		    unsigned int line = 0);

    virtual ~debug_exception();

    virtual void print(ostream& out) const;

    /* get/set the file name */ 
    virtual void set_file_name(const string& file_name) {_conf_name = file_name;}
    virtual const string& get_file_name(void) const {return _conf_name; }

 private:
    /* the source file name this exception was thrown from */
    string _conf_name;

    /* the source fun this exception was thrown from */
    string _func;

    /* the source line this exception was thrown from */
    unsigned int _line;
};

class basic_parse_exception : public basic_exception
{
 public:
    /* constructor */
    basic_parse_exception(const string& conf_file, unsigned int conf_line, 
			  const string& err);

    basic_parse_exception(const string& conf_file, unsigned int conf_line, 
			  const string& err, const token& tok);

    virtual ~basic_parse_exception();

    /* print this excpetion to ostream 'out' */
    virtual void print(ostream& out) const;
    
 private:
    
    /* did a token cause this exception? */
    bool _has_token;
    
    /* the token that caused this excpetion */
    token _t;
};

class debug_parse_exception: public basic_parse_exception
{
 public:
    debug_parse_exception(const string& conf_file, 
			  unsigned int conf_line, const string& err, 
			  const string& fname = "",
			  const string& func = "", unsigned int line = 0);

    debug_parse_exception(const string& conf_file, 
			  unsigned int conf_line, const string& err, 
			  const token& t, const string& fname = "",
			  const string& func = "", unsigned int line = 0);

    virtual ~debug_parse_exception();

    virtual void print(ostream& out) const;

 private:
    debug_exception _d;
};

class basic_validate_exception: public basic_exception
{
 public:
    /* constructor */
    basic_validate_exception(const string& conf_file, unsigned int conf_line, 
			     const string& err);

    virtual ~basic_validate_exception();

    /* print this excpetion to ostream 'out' */
    virtual void print(ostream& out) const;
};

class debug_validate_exception: public basic_validate_exception
{
 public:
    debug_validate_exception(const string& conf_file, 
			  unsigned int conf_line, const string& err, 
			  const string& fname = "",
			  const string& func = "", unsigned int line = 0);

    virtual ~debug_validate_exception();

    virtual void print(ostream& out) const;
 private:
    debug_exception _d;
};


#ifdef SCT_CONFIG_DEBUG

#define parse_exception(file, line, err) \
  debug_parse_exception(file, line, err, __FILE__, __FUNCTION__, __LINE__)

#define parse_exception_t(file, line, err, token) \
  debug_parse_exception(file, line, err, token, __FILE__, __FUNCTION__, __LINE__);

#define validate_exception(file, line, err) \
  debug_validate_exception(file, line, err, __FILE__, __FUNCTION__, __LINE__);

#else /* SCT_CONFIG_DEBUG is not defined */

#define parse_exception(file, line, err) \
  basic_parse_exception(file, line, err)

#define parse_exception_t(file, line, err, token) \
  basic_parse_exception(file, line, err, token)
 
#define validate_exception(file, line, err) \
  basic_validate_exception(file, line, err)


#endif /* SCT_CONFIG_DEBUG */  

#endif /* _PARSE_EXCEPTION_H_ */
