/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 *$Id: parser.cpp,v 1.37 2002/06/15 15:16:13 mulix Exp $
 */

/* parser implementation */

#include <string>
#include <strstream>
#include <cstdio>

#include "parser.h"
#include "parse_exception.h"
#include "token.h"
#include "validity_checks.h"
#include "logger.h"
#include "param_types.h"
#include "tokens_to_types.h"
#include "filter_parser.h"
#include "sct_rules/rule_engine_public.h"

#define WHITESPACE " \t\n\r"
static std::string whitespace_str; 

parser::parser(strstream& config_data, string file_name):
    _data(config_data), _fname(file_name), _default_log_format_before(), 
    _default_log_format_after(), _late_parsers()
{
}


/* the main entry point. performs initial processing and then parses */
/* each rule                                                         */
bool parser::parse(list<rule>& rules)
{
    /* parsing the configuration data. */

    /* strip comments from the file. we do this in a pretty bizzare way  */
    /* due problems encountered with strstreams. what we need to do here */
    /* is read all of the data from the strstream, get rid of all        */
    /* comment lines and then spill it into a new strstream and replace  */
    /* the old one.                                                      */
    string data;
    char buffer[2048];
    bool multi_line_comment = false;
     
    /* read the data line by line from the file.  */
    /* if the line is a comment dump it.  */
    while (_data.getline(buffer, sizeof(buffer))){
	/* strip c style multi line comments */
	char* comment_start;
	while (multi_line_comment || (comment_start = strstr(buffer, "/*"))){
	    if (!comment_start)
		comment_start = buffer; 

	    char* comment_end = strstr(buffer, "*/");
	    if (comment_end){
		/* get rid of the comment by copying what's beyond it over it */
		memcpy(comment_start, comment_end + 2, 
		       buffer + sizeof(buffer) - comment_end);
		multi_line_comment = false;
	    } else {
		multi_line_comment = true;
		break;
	    }
	}
	/* if this is a comment, we do an ugly hack and push in  */
	/* an empty line to get the line numbers in diagnostics  */ 
	/* correctly. not the best way to do it, but it's simple */
	if ((multi_line_comment) ||
	    (buffer[0] == '#') || 
	    ((buffer[0] == '/') && (buffer[1] == '/'))){
	    data+='\n';
	    continue;
	}

	/* add the current buffer to the data */
	data+=buffer;

	/* getline discarded the '\n' */
	data+='\n';
    }
     
    /* do the actual parsing */
    return parse_file(data, rules);
}

bool 
parser::parse_file(const string& data, list<rule>& rules)
{
    int line_num = 1;
    token t;
    string::size_type location = 0;
    /* read_token will return a token with the entire rule data contained */
    /* and a type of RULE */
    try {
	while (read_token(t, data, &location, &line_num)){
	    switch (t.get_type()) {
	    case (PR_RULE): 
	    {
		rule r;
		r.set_rule_line_num(t.get_type_line_num());
		r.set_rule_data_line_num(t.get_value_line_num());
		parse_rule(r, t.get_value());
		rules.push_back(r);
		break;
	    }
	    case (PR_LOG_FORMAT):
	    {
		parse_default_log_format(t.get_value_line_num(), t.get_value());
		break;
	    }
	    default:
		throw parse_exception_t(_fname, line_num, 
					"error in configuration file", t);
	    }
	}
    }catch (exception_intf& e){
	/* only we know the file name */
	if (e.get_file_name() == "")
	    e.set_file_name(_fname);
	log_error(e);
	return false;
    }
    return true;
}

void
parser::parse_rule(rule& r, const string& rule)
{
    token t;
    string::size_type pos = 0;
    string filter_clause;
    bool has_filter_clause = false; 
    int line_num = r.get_rule_data_line_num();
    while(read_token(t, rule , &pos, &line_num)){
	/*read tokens from the data and act based on their types. */
	switch (t.get_type()){
	case(PR_SYSCALL_NAME):{
	    r.set_syscall_identifier(t.get_value());
	    r.set_syscall_name(t.get_value());
	    break;
	}
	case(PR_RULE_NAME):{
	    r.set_rule_name(t.get_value());
	    break;
	}
	case(PR_RULE_WHEN):{
	    if (t.get_value().compare("before") == 0){
	        r.set_rule_when(RW_BEFORE);
	    }
	    else if (t.get_value().compare("after") == 0){
	        r.set_rule_when(RW_AFTER);
	    }
	    else {
		string err = "unknown 'when' value '" + t.get_value() +
			     "' in configuration file";
		throw parse_exception(_fname, line_num, err);
	    }
	    break;
	}
	case(PR_PARAM_CONDITION):{
	    param_conditional* pcond = r.new_param_conditional();
	    if (!pcond){
		throw parse_exception_t(_fname, line_num, 
				      "out of memory", t);
	    }
	    pcond->set_line_num(t.get_type_line_num());
	    pcond->set_data_line_num(t.get_value_line_num());
	    parse_param_condition(pcond, t.get_value());
	    break; 
	}
	case(PR_PROCESS_CONDITION):{
	    process_conditional* pcond = r.new_process_conditional();
	    if (!pcond){
		throw parse_exception_t(_fname, line_num, 
				      "out of memory", t);
	    }
	    pcond->set_line_num(t.get_type_line_num());
	    pcond->set_data_line_num(t.get_value_line_num());
	    parse_process_condition(pcond, t.get_value());
	    break;
	}
	case(PR_ACTION):
	{
	    register_for_delayed_parsing(&parser::parse_action, t, r);
	    break;
	}
	case(PR_CONDITIONS):{
	    r.set_condition_matching(t.get_value());
	    break;
	}
	case (PR_FILTER_EXPRESSION):{
	    /* we defer parsing the filter expression until we parse */
	    /* all of the other rule fields, in particular, the */
	    /* syscall name */
	    if (has_filter_clause)
		throw parse_exception_t(_fname, line_num, 
					"more than one filter expression in file", t);

	    has_filter_clause = true;
	    register_for_delayed_parsing(&parser::parse_filter_expression, t, r);
	    break;
	}
	default:
	    throw parse_exception_t(_fname, line_num, 
				    "invalid token parsed", t);
	}
    }

    activate_delayed_parsers();
    _late_parsers.clear();

    /* finalize the rule - will throw an exception if there's any problem */
    (void)r.finalize();

    /* validate the rule, by making sure it has all required fields and  */
    /* no collisions in the instructions. will throw an exception        */
    /* if there's any problem                                            */           
    (void)r.validate();
}

void 
parser::parse_filter_expression(const string& str, rule& r, 
				int line, void* param)
{
    filter_parser p(r.get_syscall_name());
    p.set_line_num(line);

    sct_filter_node_t* exp = NULL;
    p.parse(exp, str);
    r.set_filter_expression(exp);
}

void 
parser::parse_action(const string& str, rule& r, 
		     int line, void* param)
{
    /* what kind of action is it? */
    token t; 
    string::size_type pos = 0;
    parser_sct_action_t act_type = ACT_LAST_ACT;
    int act_error_code = 0; 
    int act_log_prio = 4; /* TODO: why? */
    string act_log_format; 
    bool has_log_format = false; 
    int pid = -1; /* for kill and suspend, -1 means 'current process' */ 
    while(read_token(t, str , &pos, &line)){
	switch (t.get_type()) 
	{
	case (PR_ACT_TYPE):
	{
	    act_type = tokens::get_action(t.get_value());
	    if (act_type == ACT_LAST_ACT)
		throw parse_exception_t(_fname, line, 
					"invalid action type parsed", t);
	    break;
	}
	case (PR_ACT_SETPARAMATTR):
	{
	    register_for_delayed_parsing(&parser::parse_act_set_param_attr, t, r);
	    break;
	}
	case (PR_ACT_ERROR_CODE):
	{
	    string err_code = t.get_value();
	    if (!is_int(err_code))
		throw parse_exception_t(_fname, line, 
					"action error code must be "
					"numeric", t)
            /* cant fail, we just checked */
	    sscanf(err_code.c_str(), "%d", &act_error_code); 
	    break;
	}
	case (PR_LOG_FORMAT):
	{
	    has_log_format = true; 
	    act_log_format = t.get_value();
	    /* make sure it has a '\n' in the end */
	    if (act_log_format != "")
		act_log_format += '\n';
	    break;
	}
	case (PR_ACT_LOG_PRIO):
	{
	    string prio = t.get_value();
	    if (!is_int(prio))
		throw parse_exception_t(_fname, line, 
					"action log priority must be "
					"numeric", t);

	    /* cant fail, we just checked */
	    sscanf(prio.c_str(), "%d", &act_log_prio); 
	    break;
	}
	case (PR_ACT_PID): /* fallthrough */
	{
	    string spid = t.get_value(); 
	    if (!is_int(spid))
		throw parse_exception_t(_fname, line, 
					"pid for kill or suspend must "
					"be numeric", t);
	    sscanf(spid.c_str(), "%d", &pid); 
	    break; 
	}
	default:
	    throw parse_exception_t(_fname, line, 
				    "unknown 'action' directive", t);
	}
    }
    
    /* we have all of the required params, let's roll */
    switch (act_type) {
    case (ACT_LOG):
    {
	string& lf = (has_log_format ? act_log_format : 
		      ((r.get_rule_when() == RW_BEFORE) ? _default_log_format_before : 
		       _default_log_format_after));
	
	r.set_action(sct_make_action_log(act_log_prio, lf.c_str()));
	break;
    }
    case (ACT_FAIL):
    {
	r.set_action(sct_make_action_fail_syscall(act_error_code));
	break;
    }
    case (ACT_KILL):
    {
	r.set_action(sct_make_action_kill_process(pid)); 
	break; 
    }
    case (ACT_SUSPEND):
    {
	r.set_action(sct_make_action_suspend_process(pid)); 
	break; 
    }
    default:
	throw parse_exception(_fname, line, 
			      "no action type given in action clause");
    }
}

void 
parser::parse_act_set_param_attr(const string& str, rule& r, 
		     		 int line, void* param)
{
    /* what kind of attr should we set? */
    token t; 
    string::size_type pos = 0;
    int var_index = -1; 
    sct_var_attr_name_t attr_name = VAN_LAST_TYPE;
    bool got_attr_val = false;
    string attr_val_str = "";
    int attr_val = 0;

    while(read_token(t, str , &pos, &line)){
	switch (t.get_type()) 
	{
	case (PR_ACT_SETPARAMATTR_ATTR_NAME):
	{
	    attr_name = tokens::get_var_attr(t.get_value());
	    if (attr_name == VAN_LAST_TYPE)
		throw parse_exception_t(_fname, line, 
					"invalid attr name parsed", t);
	    break;
	}
	case (PR_ACT_SETPARAMATTR_VAR_INDEX):
	{
	    string var_index_str = t.get_value();
	    if (!is_int(var_index_str))
		throw parse_exception_t(_fname, line, 
					"parameter index must be "
					"numeric", t)
            /* cant fail, we just checked */
	    sscanf(var_index_str.c_str(), "%d", &var_index); 
	    break;
	}
	case (PR_ACT_SETPARAMATTR_ATTR_VAL):
	{
	    attr_val_str = t.get_value();
	    if (attr_val_str[0] == '"') {
		attr_val_str = attr_val_str.substr(1, attr_val_str.length()-2);
	    }
	    if (!syscalls::has_struct(attr_val_str))
		throw parse_exception_t(_fname, line, 
					"attribute value isn't a valid "
					"struct type", t)
            /* cant fail, we just checked */
	    attr_val = syscalls::struct_type_to_index(attr_val_str);
	    got_attr_val = true;
	    break;
	}
	default:
	    throw parse_exception_t(_fname, line, 
				    "unknown 'set_param_attr' directive", t);
	}
    }
    
    /* get the 'action' object. */
    sct_action_t* action = r.get_action();
    if (!action)
	throw parse_exception(_fname, line, 
			      "internal error - couldn't find 'action' "
			      "data while handling 'set_param_attr'");
    /* make sure we got all params from the user. */
    if (var_index == -1)
	throw parse_exception(_fname, line, 
			      "no param index given in set_param_attr clause");
    if (attr_name == VAN_LAST_TYPE)
	throw parse_exception(_fname, line, 
			      "no attr name given in set_param_attr clause");
    if (!got_attr_val)
	throw parse_exception(_fname, line, 
			      "no attr value given in set_param_attr clause");

    set_action_var_indexed_attr_set_command(action,
					    VT_PARAM,
					    var_index - 1,
					    attr_name,
					    attr_val);
}

void 
parser::register_for_delayed_parsing(parse_func fn, 
				     token& t, rule& r)
{
    delayed_parser d(fn, t, r, NULL);
    _late_parsers.push_back(d);
}
				     
void 
parser::activate_delayed_parsers(void)
{
    list<delayed_parser>::const_iterator it = _late_parsers.begin();
    for ( ; it != _late_parsers.end(); ++it ) {
	const delayed_parser& d = *it;
	parse_func pfn = d._func;
	(this->*pfn)(d._tok.get_value(), d._r, 
		     d._tok.get_value_line_num(), 
		     d._param);
    }
}

void 
parser::parse_default_log_format(int line_num, const string& log_formats)
{
    token t;
    string::size_type pos = 0;
    string lf_before, lf_after;
    /* warn if the user sets a 'before' or */
    /* 'after' format, *and* a general format. */
    const unsigned int got_nothing = 0x0; 
    const unsigned int got_before = 0x1;
    const unsigned int got_after = 0x2;
    const unsigned int got_default = 0x4;
    unsigned int got = 0;
    while(read_token(t, log_formats , &pos, &line_num)){
	switch (t.get_type()){
	case(PR_LOG_FORMAT_BEFORE): {
	    lf_before = t.get_value();
	    got |= got_before;
	    break;
	}
	case(PR_LOG_FORMAT_AFTER): {
	    lf_after= t.get_value();
	    got |= got_after;
	    break;
	}
	case(PR_LOG_FORMAT_DEFAULT): {
	    /* the user must have specified the log_format  */
	    /* value directly. set before here, and we'll */
	    /* set after later */
	    lf_before = lf_after = t.get_value();
	    got |= got_default;
	    break;
	}
	default: {
	    got = got_nothing;
	    break;
	}
	}
    }
    /* what did the user specify? */
    switch (got){
    case (got_nothing): {
	strstream buf;
	buf << "log_format directive with an invalid log format" << endl << ends;
	throw validate_exception(_fname, line_num, buf.str());
	/* not reached */
	break;
    }
    case (got_before): /* fallthrough */
    case (got_after): {
	strstream buf;
	buf << "only '" << ((got & got_before) ? "before" : "after")
	    << "' log_format directive found. you must specify '"
	    << ((got & got_before) ? "after" : "before" )
	    << "' as well" << endl << ends;
	throw validate_exception(_fname, line_num, buf.str());
	/* not reached */
	break;
    }
    case ((got_before | got_after)): {
	/* we should validate it here */
	_default_log_format_before = lf_before;
	_default_log_format_after = lf_after;
	break;
    }
    case (got_default): {
	/* validate, validate, validate */
	_default_log_format_before = _default_log_format_after = lf_before;
	break;
    }
    default: { /* any other combination of default and before/after */
	strstream buf;
	buf << "log_format 'before' or 'after' specified together with "
	    "a default log_format." << endl << ends;
	throw validate_exception(_fname, line_num, buf.str());
	break;
    }
    }
    /* make sure the log formats end in '\n' */
    if (_default_log_format_before != "")
	_default_log_format_before += '\n';

    if (_default_log_format_after != "")
	_default_log_format_after += '\n';

}

static void trim(string& s)
{
#if 0
    cerr << "trim got '" << s << "'" << endl; 
#endif
    string::size_type beg, end; 
    beg = s.find_first_not_of(WHITESPACE); 
    end = s.find_last_not_of(WHITESPACE); 
    s = s.substr(beg, end - beg + 1); 
#if 0
    cerr << "trim returns '" << s << "'" << endl; 
#endif
}

bool parser::read_token(token& t,
			const string& rule,
			string::size_type* pos,
			int* p_line_num)
{
    int token_type_line_num;
    int token_value_line_num;
    /* a complex token is one which has the begin and end */
    /* ('{' and '}') chars after the name.                */
    bool complex = false;
    int complex_line_num;

    /* find the first non whitespace character. */
    string::size_type beg = rule.find_first_not_of(WHITESPACE, *pos);
    if (beg == rule.npos){
	return false;
    }

    /* update line number. */
    *p_line_num += count_lines(rule, *pos, beg);
    token_type_line_num = *p_line_num;
     
    /* find the end of this word. this is either whitespace, the '=' */
    /* sign or the '{' sign.                                         */
    string::size_type end = rule.find_first_of(WHITESPACE"={", beg+1);
    if (end == rule.npos){
	return false;
    }

     
    /* is it a complex token? */
    if (rule[end] == '{')
	complex = true;
     
    /* this word is the 'type' of the token.                 */
    /* we take the address of a temp variable here, so care  */
    /* must be taken not to use it outside of this function. */
    string tok = rule.substr(beg, end - beg);
    trim(tok); 

    /* update line number. */
    *p_line_num += count_lines(rule, beg+1, end);

    do {
	/* now find the beginning of the 'value' part of the token */
	beg = rule.find_first_not_of(WHITESPACE, end+1);
	if (beg == rule.npos){
	    return false;
	}
	/* update line number. */
	*p_line_num += count_lines(rule, end+1, beg);
	if (rule[beg] != '=')
	    break;
	end = beg; 
    } while (1); 

    /* again, check if this token is complex. */
    if (rule[beg] == '{'){
	complex = true;
	complex_line_num = *p_line_num;
	/* if it is complex, skip forward a few chars until the */
	/* actual beginning of the value.                       */
	string::size_type old_beg = beg;
	beg = rule.find_first_not_of(WHITESPACE,beg+1);
	if (beg == rule.npos)
	    return false;
        /* update line number. */
        *p_line_num += count_lines(rule, old_beg+1, beg);
    }

    token_value_line_num = *p_line_num;

    /* if it's not a complex token, the value is only one word, so */
    /* it ends with whitespace.                                    */
    if (!complex){
	end = rule.find_first_of(WHITESPACE, beg+1);
	/* no need to check for npos - this value might extend to the */
	/* end of the line */ 
        
        /* update line number. */
        *p_line_num += count_lines(rule, beg+1, end);
    }else{
	/* if it is a complex token, it might contain other complex */
	/* tokens, so we need to find the '}' sign matching ours.   */
	int count = 1;
	end = rule.find_first_of("{}", beg+1);
	while (end != rule.npos){
	    if (rule.at(end) == '{'){
		++count;
	    }else{
		--count;
	    }
	    if (count == 0)
		break;
	    end = rule.find_first_of("{}", end+1);
	}
	if (count != 0){
	    throw parse_exception(_fname, complex_line_num,
				  "mismatched brackets in input");
	}
        /* update line number. */
        *p_line_num += count_lines(rule, beg+1, end);
    }
    
    parser_token_t type = tokens::get_parser_token(tok);
    t.set_type(type);
    t.set_type_line_num(token_type_line_num);
    
    if (t.get_type() == PR_LAST_TOKEN_TYPE) { 
	string err = "unknown string '" + tok + "' in configuration file";
	throw parse_exception(_fname, token_type_line_num, err);
    }

    /* ok, trim whitespace from the end and copy the value part. */
    string tok_value = rule.substr(beg, end - beg);
    trim(tok_value); 

    t.set_value(tok_value);
    t.set_value_line_num(token_value_line_num);

    *pos = (end == string::npos ? string::npos : end + 1);
    return true;
}

void parser::parse_param_condition(
    param_conditional* pcond, const string& data)
{
    token t;
    string::size_type pos = 0;
    int line_num = pcond->get_data_line_num();
    
    while (read_token(t, data, &pos, &line_num)){
	switch(t.get_type()){
	case(PR_PARAM_LOCATION):{
	    pcond->set_location(t.get_value());
	    break;
	}
	case(PR_PARAM_LOCATION_FIELD):{
	    pcond->set_location_field(t.get_value());
	    break;
	}
	case(PR_OPERATOR):{
	    pcond->set_operator(t.get_value());
	    break;
	}
	case(PR_VALUE):{
	    pcond->set_value(t.get_value());
	    break;
	}
	default:
	    throw parse_exception_t(_fname, line_num, 
				  "unknown token in parameter condition", t);
	}
    }
}

void parser::parse_process_condition(
    process_conditional* pcond, const string& data)
{
    token t;
    string::size_type pos = 0;
    int line_num = pcond->get_data_line_num();

    while (read_token(t, data, &pos, &line_num)){
	switch(t.get_type()){
	case(PR_PROCESS_FIELD):{
	    pcond->set_process_field(t.get_value());
	    break;
	}
	case(PR_OPERATOR):{
	    pcond->set_operator(t.get_value());
	    break;
	}
	case(PR_VALUE):{
	    pcond->set_value(t.get_value());
	    break;
	}
	default:
	    throw parse_exception_t(_fname, line_num, 
				    "unknown token in process condition", t);
	}
    }
}

void parser::log_error(const exception_intf& e)
{
    strstream buf; 
    e.print(buf);
    buf << endl << ends;
    
    logger::log(logger::LOG_LEVEL_ERROR, "%s", buf.str());
}

/* count the number of lines (new-line chars) in the given string, */
/* between 'beg' and 'end'.                                        */
int parser::count_lines(const string& data,
			 string::size_type beg,
			 string::size_type end)
{
    int line_span = 0;
    string::size_type curr = data.find_first_of("\n", beg);
    if (end > data.length())
	end = data.length();

    while (curr <= end) {
	line_span++;
        curr = data.find_first_of("\n", curr+1);
    }

    return line_span;
}

