/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 * $Id: parser.h,v 1.14 2002/06/15 15:16:13 mulix Exp $
 */

/*  parser is responsible of parsing the configuration file, as defined in 
 *  the config.file document. it then builds a list of rules which can
 *  sent to the module
 *
 *  @author: mulix@actcom.co.il
 */

#ifndef _PARSER_H_
#define _PARSER_H_

#include <string>
#include <strstream>

#include "token.h"
#include "rule.h"
#include "process_conditional.h"
#include "param_conditional.h"
#include "parse_exception.h"

using namespace std;

class parser
{
 public:
    /* constructor.                                           */
    /* @param: the configuration data, passed as a strstream  */
    parser(strstream& config_data, string file_name);

    /* parse is the one and only entry point into the parser.  */
    /* if the parsing failed the function will return it.      */
    /* if the parsing of any rule is successful, we return the */
    /* parsed rules in the rules list.                         */
    /* returns true if parsing of ALL rules was successful.    */
    bool parse(list<rule>& rules);

    inline const string& get_file_name() const { return _fname; } 

    /* count the number of lines (new-line chars)  in the */
    /* given string, between 'beg' and 'end'.             */
    int count_lines(const string& data,
		    string::size_type beg,
		    string::size_type end);
    
 private:
    /* parse_rule accepts a string of rule data and parses it into a */
    /* rule structure.                                               */
    void parse_rule(rule& r, const string& data);

    /* parse_log_format accepts a string containing log_format information, */
    /* parses it and stores it in log_format_before, log_format_after or    */
    /* both, as applicable. if an error occurs, an exception is thrown.     */
    void parse_default_log_format(int line_num, const string& log_formats);

    /* read_token does most of the parsing. it accepts a string of rule */
    /* data, a position to start parsing from, and a line number to     */
    /* start counting from.                                             */
    /* it returns in its arguments a token class with data about the    */
    /* parsed token, the position to start the next parsing from, and   */
    /* the line number of the end of the token.                         */
    /* returns false when there are no more tokens to be read           */
    /* throws a parse_exception if encounters an unknown token          */
    bool read_token(token& t,
		    const string& rule,
		    string::size_type* pos,
		    int* p_line_num);

    /* parse_process_condition accepts a process_conditional structure */
    /* and a string of rule data. it parses the data and fills the     */
    /* process conditional.                                            */
    void parse_process_condition(process_conditional* pcond, const string& data);

    /* parse_process_condition accepts a param_conditional structure */
    /* and a string of rule data. it parses the rule data and fills  */
    /* the param_conditional structure.                              */
    void parse_param_condition(param_conditional* pcond, const string& data);

    /* log_error accepts a parse_exception and logs it to the log file */
    void log_error(const exception_intf& e);
     
    /* members */
    strstream& _data;

    string _fname;
    unsigned int _line;

    string _default_log_format_before;
    string _default_log_format_after;

    typedef void (parser::* parse_func)(const string& parse_me, rule& r, 
					int line_num, void* param);
    class delayed_parser {
    public:
	delayed_parser(parse_func fn, const token& tok,
		       rule& r, void* param): 
	    _func(fn), _tok(tok), _r(r), _param(param) {};
	
	parse_func _func;
	token _tok;
	rule& _r; 
	void* _param;
    };
    
    std::list<delayed_parser> _late_parsers;

    bool parse_file(const string& data, list<rule>& rules);

    void parse_filter_expression(const string& str, rule& r, 
				 int line, void* param);

    void parse_action(const string& str, rule& r, 
		      int line, void* param);

    void parse_act_set_param_attr(const string& str, rule& r, 
		      		  int line, void* param);

    void register_for_delayed_parsing(parse_func fn, 
				      token& t, rule& r);
    
    void activate_delayed_parsers(void);
};

#endif /* _PARSER_H_ */
