/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 *$Id: rule.cpp,v 1.42 2002/06/15 15:16:13 mulix Exp $
 */

#include <errno.h>
#include <iostream>
#include <list>
#include <strstream>

#include "parser.h"
#include "parse_exception.h"
#include "rule.h"
#include "validity_checks.h"
#include "logger.h"
#include "syscalls.h"
#include "token.h"
#include "tokens_to_types.h"

#include "sct_rules/rule_engine_public.h"

rule_identifier rule::last_rule_id = 0;

ostream& 
operator<<(ostream& os, const rule& r)
{
    os<<"rule: "<<r._rule_id<<endl;
    os<<"\twhen: "<<(r._rule_when == RW_BEFORE ? "before" : "after")<<endl;
    os<<"\tsyscall: "<<r._syscall_id<<endl;
    list<conditional_intf*>::const_iterator it = r._conds.begin();
    for(; it != r._conds.end(); ++it) {
	os<<"\t";
	(*it)->print(os);
    }
   
    /* only print condition matching if we have more than 1 condition clause */
    if (r._conds.size() > 1)
	os<<"\tcondition matching: "<<
	    tokens::get_condition_matching_str(r._cond_match)<<endl;

    if (r._filter_expr)
	sct_print_filter_node(r._filter_expr);

    char* desc = sct_get_action_as_string(r._act);
    os<<"\taction: "<< (desc ?: "(nil)") << endl;
    free(desc); /* allocated using malloc */
    return os;
    
}

rule::rule():_rule_id(++last_rule_id),_rule_when(RW_BEFORE),
	     _syscall_id(-1), _syscall_name(), 
	     _act(NULL), _cond_match(COND_LAST_COND), 
	     _filter_expr(NULL)
{
}

rule::rule(const rule& r)
{
    _rule_id = r._rule_id;
    _rule_when = r._rule_when;
    _syscall_id = r._syscall_id;
    _rule_name = r._rule_name;
    _cond_match = r._cond_match;
    // shallow copy the lists of pointers
    list<conditional_intf*>::const_iterator it = r._conds.begin();
    for (; it != r._conds.end(); ++it){
	_conds.push_back(*it);
    }
    // get the barf bag ready
    // we transfer ownership in of the pointers, instead of copying them, 
    // and i no longer remember why. FIX THIS. 
    ((rule&)r)._conds.clear();
    _filter_expr = r._filter_expr;
    _act = r._act;
    ((rule&)r)._filter_expr = NULL;
    ((rule&)r)._act = NULL;
}

rule::~rule()
{
    list<conditional_intf*>::iterator it = _conds.begin();
    for (; it != _conds.end(); ++it){
	delete *it;
    }
    if (_filter_expr)
	sct_free_filter_node(_filter_expr);
}


void
rule::set_syscall_identifier(const string& str)
{
    int sid = syscalls::syscall_id(str); 
    if (sid == -1) {
	string err = "'" + str + "' is not a valid syscall id";
	throw parse_exception("", get_rule_line_num(), err);
    }
    _syscall_id = sid;
}

param_conditional* 
rule::new_param_conditional()
{
    param_conditional* pcond = new param_conditional();
    _conds.push_back(pcond);
    return pcond;
}

process_conditional* 
rule::new_process_conditional()
{
    process_conditional* pcond = new process_conditional();
    _conds.push_back(pcond);
    return pcond;
}

void 
rule::set_condition_matching(const string& str){
    _cond_match = tokens::get_condition_matching(str);
    if (_cond_match == COND_LAST_COND){
	string err = "'" + str + "' is not a valid condition matching directive";
	throw parse_exception("", get_rule_line_num(), err);
    }
}

bool
rule::validate() const
{
    bool valid = true;
    string rname = (_rule_name.length() ? _rule_name : "(unknown)");
    if (_syscall_id == 0) {
	string err = "rule '" + rname + "' does not have a valid syscall";
	throw validate_exception("", _rule_line_num, err);
    }
#if 0 /* disable the rule name check - is it really necessary? */ 
    if (_rule_name.empty()){
	strstream buf;
	buf << "rule '" << rname << "' with id '"
	    << _rule_id << "' has no rule name " << ends;
	throw validate_exception("", _rule_line_num, buf.str());
    }
#endif 
    if (!_act){
	string err = "rule '" + rname + "' does not have a valid action";
	throw validate_exception("", _rule_line_num, err);
    }
    if (!conds_valid()){
	valid = false;
    }
    if (_conds.size() > 1) {
	if (_cond_match == COND_LAST_COND) {
	    string err = "rule '" + rname + 
		"'has at least two conditional clauses "
		"and no CONDITIONAL_MATCHING clause";
	    throw validate_exception("", _rule_line_num, err);
	}
    }
    // special check for exit(1), which cannot have an 'after' rule
    // figure out a better way to do this
    if (_syscall_id == syscalls::syscall_id("exit")
	&& _rule_when == RW_AFTER) {
	    strstream err; 
	    err << "rule '" << rname << "': the exit(1) system "
		    "call is not allowed to have after rules." 
		<< ends; 
	    throw validate_exception("", _rule_line_num, err.str()); 
    }
    return valid;
}
     
bool
rule::conds_valid() const
{
    if (_filter_expr && !_conds.empty()) {
	strstream err;
	err << "rule contains both a filter expression and one "
	    "(or more) conditional clauses" << ends;
	throw validate_exception("", _rule_line_num, err.str());
    }
    bool res = true;
    list<conditional_intf*>::const_iterator it = _conds.begin();
    for (; it != _conds.end();++it){
	(void)((*it)->validate(_syscall_name));
    }
    return res;
}

void
rule::set_filter_expression(sct_filter_node_t* expr)
{
    if (expr) {
	_filter_expr = expr;
	if (!_conds.empty()) {
	    strstream err;
	    err << "rule contains both a filter expression and one "
		"(or more) conditional clauses" << ends;
	    throw validate_exception("", _rule_line_num, 
				     err.str());
	}
    }
}

static sct_filter_node_t*
add_to_filter_chain(sct_filter_node_t* head, sct_filter_node_t* link, 
		    condition_matching_t cond);

// transform this rule into a filter_node 
sct_filter_node_t* rule::filter_node() const
{
    /* if we have a filter expression, return it */
    if (_filter_expr)
	return _filter_expr;

    /* if not, build it from the conditional match clauses */
    sct_filter_node_t* all = NULL;
    /* build the differet params we want to match against */
    list<conditional_intf*>::const_iterator it1 = _conds.begin();
    for (; it1 != _conds.end(); ++it1)
    {
	sct_filter_node_t* t2 = (*it1)->sct_get_filter_node();

	/* we have the operator and two arguments, now add them to the chain itself */
	all = add_to_filter_chain(all, t2, _cond_match);
    }

    /* handle the case where we have an unconditional rule */
    if (!all) 
	all = sct_make_bool_val(1);

    if (logger::get_verbosity_level() <= logger::LOG_LEVEL_INFO)
        sct_print_filter_node(all);

    return all;
}

sct_filter_node_t*
add_to_filter_chain(sct_filter_node_t* head, sct_filter_node_t* link, 
		    condition_matching_t cond)
{
    if (!head)
	return link;

    if (!link)
	return head;

    if (cond == COND_ALL) {
	return sct_make_logical_and(head, link);
    } else if (cond == COND_ANY) {
	return sct_make_logical_or(head, link);
    } else /* COND_LAST_COND */
	return NULL;
}

bool
rule::finalize()
{
    list<conditional_intf*>::iterator it = _conds.begin();
    for ( ; it != _conds.end(); ++it ) {
	if (!(*it)->finalize(_syscall_name))
	    return false; 
    }
    return true; 
}
