/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/* 
 * $Id: syscalls.cpp,v 1.21 2002/06/15 15:16:13 mulix Exp $
 */ 

/* parse the syscalls.dat file and make it available */
/* to the parser, for validity checks                */  

#include <stdio.h>
#include <iostream>
#include <fstream>
#include <list>
#include <pair.h>
#include <string>
#include <algorithm>

#include "syscalls.h"
#include "validity_checks.h"
#include "logger.h"
#include "tokens_to_types.h"

syscalls::syscalls_set syscalls::_sys;
syscalls::structs_set syscalls::_structs;

/* defined in main.cpp */
extern int debug;

syscalls::syscalls()
{
}

syscalls::~syscalls()
{
}

class eNoToken{};

/* get_syscall_token will get a token out of the string, from position 
   'start' to the first occurence of any of the delimiters in 'delim'. 
   returns: the token. 'start' now points to the beginning of the next token 
   (NOT including whitespace). 
   if no token is found, an eNoToken exception is thrown. */
static string 
get_syscall_token(const string& str, string::size_type& start, 
		  string::size_type& end, const char* delim)
{
    assert(delim);

    end = str.find_first_of(delim, start); 
    if (end == str.npos)
	throw eNoToken();
    
    string token = str.substr(start, end - start);
    
    /* skip whitspace */
    start = str.find_first_not_of(" \t", end + 1);

    return token;
}

static 
void parse_params(const string& line, string::size_type pos, 
		  syscalls::param_vec& params)
{
    string::size_type p1 = pos;
    string::size_type p2;
    string match, type, name; 
    bool is_struct, is_ptr, is_array;
    int array_len;
    do {
	try {
	    /* get the matchable qualifier */
	    match = get_syscall_token(line, p1, p2, " \t,");
	    
	    /* get the data type */
	    type = get_syscall_token(line, p1, p2, " \t,");
	    
	    /* get the param name */
	    name = get_syscall_token(line, p1, p2, " \t,)");
	} catch (eNoToken& e){
	    /* TODO: report error? */
	    return ;
	}

#if 0
	cout << "parse_params: type '" << type.c_str() << "', name '"
		<< name.c_str() << "', match '" << match.c_str() << "'" << endl;
#endif

	/* insert the new param to the vector,
	   if it's matchable, give it its attributes
	   if not, mark it as an 'invalid' param */
	if (match != "matchable"){
	    name = "INVALID";
	}
	
	/* check if this is a struct or a native type. */
	is_struct = false;
	is_ptr = false;
	is_array = false;
	array_len = -1;
	sct_data_type_t dt;
    	string struct_type;
	const string struct_prefix = "struct_";
	const string const_struct_prefix = "const_struct_";
	const string array_prefix = "array[";
	const string ptr_prefix = "ptr_";
	if (type.find(struct_prefix) == 0) {
	    is_struct = true;
    	    struct_type = type.substr(struct_prefix.length(),
				      type.length() - struct_prefix.length());
	} else if (type.find(const_struct_prefix) == 0) {
	    is_struct = true;
    	    struct_type =
		type.substr(const_struct_prefix.length(),
		      	    type.length() - const_struct_prefix.length());
	} else if (type.find(array_prefix) == 0) {
	    is_array = true;
	    std::string::size_type len_beg = type.find_first_of('[') + 1;
	    std::string::size_type len_end = type.find_first_of(']', len_beg);
#if 0 
	    cout << "  array, len_beg (" << len_beg << ")" << endl;
	    cout << "  array, len_end (" << len_end << ")" << endl;
#endif 
	    if (len_end == type.npos)
		return;
	    std::string array_len_str = type.substr(len_beg, len_end-len_beg);
	    if (!is_int(array_len_str))
	        return;
	    array_len = atoi(array_len_str.c_str());
	    std::string real_type = type.substr(len_end+2);
	    dt = tokens::get_data_type(real_type);
#if 0 
	    cout << "  array, len '" << array_len_str << "' (" << array_len << ")" << endl;
	    cout << "  array, type '" << real_type << "'" << endl; 
#endif 
	    if (dt == DT_LAST_TYPE)
	        return;
	} else if (type.find(ptr_prefix) == 0) {
	    // TODO - do we realy have pointers in arrays? for now, lets
	    // consider this to be an error.
	    return;
	} else {
	    dt = tokens::get_data_type(type);
	    if (dt == DT_LAST_TYPE)
	        return;
	}

	syscalls::param* p = new syscalls::param; 
	p->name = name; p->is_struct = is_struct;
	p->is_ptr = is_ptr; p->is_array = is_array; p->array_len = array_len;
	if (is_struct) {
	    p->struct_type = struct_type;
	} else {
	    p->type = dt;
	}

	params.push_back(*p);
	
    } while (line[p2] != ')');
}

static bool get_mux_id(const string& sid, int& syscall, int& func)
{
    if (sscanf(sid.c_str(), "%d.%d", &syscall, &func) != 2)
	return false; 
    return true; 
}

bool
syscalls::load(string file_name)
{
    syscalls::_sys.clear();
    syscalls::_structs.clear();
    int struct_type_index = 0;

    ifstream file(file_name.c_str());

    char buf[2048];
    string::size_type p1 = 0, p2 = 0;
    while (file.getline(buf, sizeof(buf))) {
	if (buf[0] == '#')
	    continue;
	if (!strncmp(buf, "asis", 4))
	    continue;
	if (!strncmp(buf, "include", 7))
	    continue;
	/* create a string object, for easier handling */
	string line(buf);
	if (!strncmp(buf, "struct", 6)) {
#if 0
	    cout << endl << "struct: ";
#endif
	    /* handle 'struct' lines. */
	    p1 = line.find_first_of(" \t");
	    if (p1 == line.npos)
		continue;
	    /* find struct type. */
	    p1 = line.find_first_not_of(" \t", p1+1);
	    if (p1 == line.npos)
		continue;
	    string* struct_type = new string;
	    try {
	        (*struct_type) = get_syscall_token(line, p1, p2, " \t(");
	    } catch (eNoToken& e){
	        continue;
	    }
#if 0
	    cout << "'" << struct_type->c_str() << "':" << endl;;
#endif
	    /* we have a name, now let's get the parameters */
	    /* TODO: we allocate these on the heap instead of 
	       on the stack because of a weird bug with gcc2.91.6.
	       if allocating on the stack, the assignment of 
	       id.params = params crashes. no idea why. */
	    param_vec* params = new param_vec;
	    parse_params(line, p1, *params);

	    struct_info* info = new struct_info; 
	    info->struct_type = *struct_type;
	    info->struct_type_index = struct_type_index++;

	    /* handle struct-within-struct by expanding the 'param_vec'    */
	    /* to contain all the fields of the sub-param, in the matching */
	    /* loaction.                                                   */
	    param_vec* expanded_params = new param_vec;
	    for (unsigned int i=0; i < params->size(); i++) {
		param& a_param = (*params)[i];
		if (a_param.is_struct) {
		    /* locate the struct, and add all its fields. */
		    if (!has_struct(a_param.struct_type)) {
			continue;
		    }
		    param_vec& s_params = struct_params(a_param.struct_type);
		    for (unsigned int j=0; j < s_params.size(); j++) {
		        param s_param = s_params[j];
			string* old_name = new string(s_param.name);
			s_param.name = a_param.name + "." + *old_name;
		        expanded_params->push_back(s_param);
		    }
		}
		else {
		    expanded_params->push_back(a_param);
		}
	    }

	    info->params = *expanded_params;
	
	    _structs[*struct_type] = info;
	    continue;
	}
#if 0
        cout << endl << "syscall: ";
#endif
	/* now start parsing out the syscall id */
	p1 = line.find(':');
	if (p1 == line.npos)
	    continue;
	p1 = line.find_first_not_of(" \t", p1+1);
	if (p1 == line.npos)
	    continue;
	string* sid = new string;
	try {
	    *sid = get_syscall_token(line, p1, p2, ":");
	} catch (eNoToken& e){
	    continue;
	}
#if 0
        cout << "sid '" << sid->c_str() << "': ";
#endif
	int id = -1; 
	if (is_int(*sid)) {
	    id = atoi((*sid).c_str());
	} else { // first try a mux id, then abort 
	    int syscall, func; 
	    if (get_mux_id(*sid, syscall, func)) {
		/*  check if the syscall id is a in the id.func format */ 
		/*  we use for mux syscalls */ 
		id = MAKE_MULTIPLEX_SYSCALL_ID(syscall, func); 
	    } else {
		continue; 
	    }
	}
    
	/* good, we got the id, now the name */
	string* name = new string; 
	try {
	    *name = get_syscall_token(line, p1, p2, " \t(");
	} catch (eNoToken& e){
	    continue;
	}
#if 0
        cout << "name '" << name->c_str() << "': " << endl;
#endif
	
	/* we have a name, now let's get the parameters */
	/* TODO: we allocate these on the heap instead of 
	   on the stack because of a weird bug with gcc2.91.6.
	   if allocating on the stack, the assignment of 
	   id.params = params crashes. no idea why. */
	param_vec* params = new param_vec;
	parse_params(line, p1, *params);

	syscall_info* info = new syscall_info; 
	info->id = id; info->params = *params;
	
	_sys[*name] = info;

    }
    if (debug) {
	logger::log(logger::LOG_LEVEL_DEBUG, "syscalls.dat:\n");
	print_structs();
	print_syscalls();
	logger::log(logger::LOG_LEVEL_DEBUG, "-----------------------\n");
    }

    return (!syscalls::_sys.empty());
}

/* get the given parameter of this syscall */
/* returns NULL if no such syscall or loc  */
/* out of range. loc index is 1 based      */
const syscalls::param*
syscalls::syscall_param_at(const string& name, unsigned short int loc)
{
    /* is there such a syscall? */
    if (!has_syscall(name)) {
        return NULL;
    }

    const param_vec& params = syscall_params(name);
    if ((loc == 0) || (loc > params.size())) {
	return NULL;
    }

    const param& p = params[loc - 1];

    return &p;
}

/* get the given parameter of this struct */
/* also set the location of this field in */
/* the struct, into 'field_loc'.          */
/* returns NULL if no such struct or no   */
/* such field in struct.                  */
const syscalls::param*
syscalls::struct_param_at(const string& struct_type,
			  const string& field,
			  int* field_loc)
{
    /* is there such a struct type? */
    if (!has_struct(struct_type)) {
        return NULL;
    }

    const param_vec& params = struct_params(struct_type);

    int field_index = 0;
    param_vec::const_iterator it = params.begin();
    for ( ; it != params.end(); field_index++,it++) {
        if ((*it).name == field)
            break;
    }
    if (it == params.end()) {
	return NULL;
    }
    if (field_loc)
	*field_loc = field_index;
    const param& p = (*it);

    return &p;
}

typedef std::pair<std::string, syscalls::syscall_info*> syscall_pair;
typedef std::list<syscall_pair> syscall_info_list; 

static bool syscall_id_lt(syscall_pair& p1, syscall_pair& p2)
{
	int sid1, sid2, fid1, fid2; 
	sid1 = SYSCALL_PART_OF_SYSCALL_ID(p1.second->id); 
	sid2 = SYSCALL_PART_OF_SYSCALL_ID(p2.second->id); 
	if (sid1 < sid2)
		return true; 
	
	if (sid1 > sid2)
		return false; 
	
	fid1 = FUNC_PART_OF_SYSCALL_ID(p1.second->id); 
	fid2 = FUNC_PART_OF_SYSCALL_ID(p2.second->id); 
	if (fid1 < fid2)
		return true; 

	return false; 
}

/* we use this class to extract the info from the _sys map to a list */ 
class extract_syscall_info {
	syscall_info_list& lst; 
public:
	extract_syscall_info(syscall_info_list& x):lst(x){} 
	void operator()(const syscall_pair& i)
		{lst.push_back(i);}
};

void syscalls::print_syscalls()
{
    char sid[32]; 

    /* sort the syscalls into this list */ 
    syscall_info_list lst; 
    for_each(_sys.begin(), _sys.end(), extract_syscall_info(lst)); 
    lst.sort(syscall_id_lt); 

    syscall_info_list::const_iterator it = lst.begin();
    for ( ; it != lst.end(); ++it){
	const string& name = (*it).first;
	int id = (*it).second->id;
	sct_get_syscall_id_as_str(id, sid, sizeof(sid)); 
	const param_vec& params = (*it).second->params;
	param_vec::const_iterator p = params.begin(); 
	string params_string;
	for ( ; p != params.end(); ++p){
	    params_string += ((*p).name + ":");
	    if ((*p).is_struct) {
	        params_string += "struct ";
	        params_string += (*p).struct_type;
	    }
	    else {
		params_string += tokens::get_data_type_str((*p).type);
	    }
	    params_string += ", ";
	}
	/* remove trailing ', ', if we have it */
	if (params_string.size() >= 2)
	    params_string[params_string.size() - 2] = '\0';
	logger::log(logger::LOG_LEVEL_DEBUG, 
		    "%s : %s(%s)\n", sid, name.c_str(), 
		    params_string.c_str());  
    }
}

void
syscalls::print_structs()
{
    syscalls::structs_set::const_iterator it = _structs.begin();
    for ( ; it != _structs.end(); ++it){
	const string& struct_type = (*it).second->struct_type;
	const param_vec& params = (*it).second->params;
	param_vec::const_iterator p = params.begin(); 
	string params_string;
	for ( ; p != params.end(); ++p){
	    params_string += ((*p).name + ":" + 
			      tokens::get_data_type_str((*p).type)); 
	    if ((*p).is_array) {
		params_string += '[';
		char buf[40];
		sprintf(buf, "%d", (*p).array_len);
		params_string += buf;
		params_string += ']';
	    }
	    params_string += ", ";
	}
	/* remove trailing ', ' */
	params_string[params_string.size() - 2] = '\0';
	logger::log(logger::LOG_LEVEL_DEBUG, 
		    "struct %s(%s)\n",
		    struct_type.c_str(), 
		    params_string.c_str());  
    }
}

bool 
syscalls::has_syscall(const string& name)
{
    return (syscalls::_sys.find(name) != _sys.end());
}

bool 
syscalls::has_struct(const string& struct_type)
{
    return (syscalls::_structs.find(struct_type) != _structs.end());
}

    /* return the index of the struct with the given type, '-1' if not found. */
int
syscalls::struct_type_to_index(const string& type)
{
    int index = -1;

    structs_set::iterator it = syscalls::_structs.find(type);
    if (it != syscalls::_structs.end()) {
	index = (*it).second->struct_type_index;
    }

    return index;
}

int 
syscalls::syscall_id(const string& name)
{
    syscalls_set::const_iterator it = syscalls::_sys.find(name);
    if (it == syscalls::_sys.end())
	return -1;

    return (*it).second->id;
}

const syscalls::param*
syscalls::get_final_syscall_param(const string& sys_name, 
				  unsigned int location, 
				  const string& field, 
				  int* field_index)
{
    const syscalls::param* p = syscalls::syscall_param_at(sys_name, location);
    if (!p)
	return NULL;

    // handle structs
    if (p->is_struct) {
	const string& struct_type = p->struct_type;
        /* is there such a struct type? */
        if (!syscalls::has_struct(struct_type)) {
	    return NULL;
        }

        /* no such parameter? */
	const syscalls::param* p2 =
	    syscalls::struct_param_at(struct_type, field, field_index);

	return p2;
    }
    else
	if (field_index)
	    *field_index = -1;

    return p;
}

