/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: dev_file_private.c,v 1.12 2002/06/15 15:16:13 mulix Exp $
 */

#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "sct_ctrl_lib.h"
#include "sct_ctrl_lib_private.h"

#include "sct_rules/membuf.h"
#include "sct_rules_module/module_interface.h"

static unsigned int 
get_next_packet_id(void)
{
	static unsigned int packet_id = 0;
	if (packet_id == 0) {
		srand(time(NULL)^getpid());
		packet_id = rand();
	} else
		++packet_id;
	
	return packet_id;
}

static const char* 
get_device_file_name(sct_ctrl_t ctrl)
{
	char* name;

	assert(ctrl);

	name =  (char*)get_user_data(ctrl);

	/* sane default, for now. change it later */
	return (name ? name : "/dev/sct_ctrl");
}

const char* 
resp_result_to_string(int res)
{
	switch (res){
	case CTL_TRACKER_SUCCESS:          return "the operation was succesfull";
	case CTL_TRACKER_EINVAL:           return "the request packet is invalid";
	case CTL_TRACKER_ENOMEM:           return "the kernel could not allocate memory";
	case CTL_TRACKER_EINTR:	           return "uperforming the request was interrupted - "
						   "the client should try again";
	case CTL_TRACKER_ENOSYS:	   return "no such syscall (invalid syscall_id parameter)";
	case CTL_TRACKER_ESYSUNSUP:        return "this syscall is not supported at the moment";
	case CTL_TRACKER_ENOENT:	   return "no such tracking rule";
	case CTL_TRACKER_EBADTRACKINGRULE: return "the tracking rule could not be deserialized";
	default:                           return "unknown response";
	}
}

void 
destroy_request(dev_req_t* req)
{
	/* free the data field, if we have it */ 
	if (req && req->data)
		free(req->data); 
}

static int 
dev_file_open(const char* dev_file)
{
	assert(dev_file);

	return open(dev_file, O_RDWR);
}

static int 
dev_file_close(int fd)
{
	return close(fd);
}

static ssize_t
dev_file_read(int fd, void* buf, size_t sz)
{
	size_t bytes_read = 0;
	ssize_t rc = 0;
	while (bytes_read < sz){
		if ((rc = read (fd, (unsigned char*)buf+bytes_read, 
				sz-bytes_read)) == -1){
			if ((errno == EINTR) || (errno == EAGAIN))
				continue;
			break;
		} else if (rc == 0) /* eof */
			break;
		bytes_read += rc;
	}
	return (rc == -1 ? -1 : bytes_read);
}

static ssize_t
dev_file_write(int fd, void* buf, size_t sz)
{
	size_t bytes_written = 0;
	ssize_t rc = 0;
	while (bytes_written < sz){
		if ((rc = write (fd, (unsigned char*)buf+bytes_written, 
				 sz-bytes_written)) == -1){
			if ((errno == EINTR) || (errno == EAGAIN))
				continue;
			break;
		}
		bytes_written += rc;
	}
	return (rc == -1 ? -1 : bytes_written);
}

/* update the size field of this request */ 
void 
update_size(dev_req_t* req)
{
	assert(req); 
	if (!req)
		return; 

	req->size = sizeof(req->size) + 
		sizeof(req->id) + 
		sizeof(req->cmd) +
		sizeof(req->result) + 
		sizeof(req->data_sz) + 
		req->data_sz + 
		sizeof(req->magic);
}

void
build_request(dev_req_t* req, unsigned int cmd,
	      int rule_id, sct_rule_when_t when, int syscall_id, 
	      unsigned char* buf, unsigned int length, int enable_rule)
{
	assert(req);

	memset(req, 0, sizeof(*req));

	req->id = get_next_packet_id();
	req->cmd = cmd;
	req->data_sz = length;
	req->data = buf;
	req->magic = DEV_FILE_MAGIC_COOKIE;

	update_size(req); 
}

static int 
send_request(int fd, dev_req_t* req)
{
	sct_membuf_t* mbuf; 
	int rc = 0; 
	size_t sz; 

	assert(req); 
	
	/* serialize the request into the membuf */ 
	if (!(mbuf = sct_membuf_create(req->size, NULL, NULL)))
		goto done; 

	if (sct_membuf_add_chunk(mbuf, &req->size, sizeof(req->size)) < 0)
		goto free_mbuf; 

	if (sct_membuf_add_chunk(mbuf, &req->id, sizeof(req->id)) < 0)
		goto free_mbuf; 

	if (sct_membuf_add_chunk(mbuf, &req->cmd, sizeof(req->cmd)) < 0)
		goto free_mbuf; 

	if (sct_membuf_add_chunk(mbuf, &req->result, sizeof(req->result)) < 0)
		goto free_mbuf; 

	if (sct_membuf_add_chunk(mbuf, &req->data_sz, sizeof(req->data_sz)) < 0)
		goto free_mbuf; 

	if (sct_membuf_add_chunk(mbuf, req->data, req->data_sz) < 0)
		goto free_mbuf; 

	if (sct_membuf_add_chunk(mbuf, &req->magic, sizeof(req->magic)) < 0)
		goto free_mbuf; 

	sz = sct_membuf_get_serialized_size(mbuf); 
	rc = (dev_file_write(fd, sct_membuf_get_data(mbuf), sz) == sz); 
		
 free_mbuf:
	sct_membuf_destroy(mbuf); 
 done:
	return rc; 
}

static int
recv_response(int fd, dev_req_t* resp)
{
	unsigned char buf[16 * 1024];
	unsigned char* whead;
	int packetsz, rc;
	sct_membuf_t* mbuf; 
	
	assert(fd >= 0 && resp);
	
	rc = dev_file_read(fd, buf, sizeof(buf));

	/* we must read *at least* the size of the structure's size field on the first read */ 
	if (rc < sizeof(resp->size)) {
		fprintf(stderr, "%s: failed to read from the device the size of "
			"the response, got '%d'\n", __FUNCTION__, rc);
		return 0;
	}
		
	packetsz = rc;
	
#if 0
	fprintf(stderr, "read from the kernel packet size: %d\n", packetsz); 
#endif /* 0 */ 
	
	memcpy(&resp->size, buf, sizeof(resp->size));
	whead = buf + sizeof(resp->size);
	
	/* check that the reported packet size is equal to the */
	/* ammount of bytes we actually read from the kernel */
	/* TODO: if not, try reading again? */ 
	if (resp->size != packetsz) {
		fprintf(stderr, "%s: read '%d' from the device, but "
			"packet size is '%d'\n", __FUNCTION__, 
			packetsz, resp->size);
		return 0;
	}

	/* create a membuf with this buffer, minus the */
        /* size we've already read from it */ 
	if (!(mbuf = sct_membuf_create_with_data(whead, 
						 packetsz - sizeof(resp->size),
						 NULL, NULL))) {
		fprintf(stderr, "%s: creaing the membuf failed?!\n", 
			__FUNCTION__); 
		return 0; 
	}
	
	if (sct_membuf_get_chunk(mbuf, &resp->id, sizeof(resp->id)) 
	    != sizeof(resp->id))
		goto free_mbuf; 

	if (sct_membuf_get_chunk(mbuf, &resp->cmd, sizeof(resp->cmd)) 
	    != sizeof(resp->cmd))
		goto free_mbuf; 

	if (sct_membuf_get_chunk(mbuf, &resp->result, sizeof(resp->result))
	    != sizeof(resp->result))
		goto free_mbuf; 

	if (sct_membuf_get_chunk(mbuf, &resp->data_sz, sizeof(resp->data_sz))
	    != sizeof(resp->data_sz))
		goto free_mbuf; 

	/* if we got data, we need to allocate some space for it */ 
	if (resp->data_sz) {
		unsigned char* tmp = (unsigned char*)malloc(resp->data_sz); 
		if (!tmp)
			goto free_mbuf; 

		resp->data = tmp; 
	}

	if (sct_membuf_get_chunk(mbuf, resp->data, resp->data_sz) != resp->data_sz)
		goto free_data; 

	if (sct_membuf_get_chunk(mbuf, &resp->magic, sizeof(resp->magic)) 
	    != sizeof(resp->magic))
		goto free_data; 

	sct_membuf_destroy(mbuf); 
	return 1; 
	
 free_data: 
	free(resp->data); 
	resp->data = NULL; 
 free_mbuf: 
	sct_membuf_destroy(mbuf); 
	return 0; 
}

static int 
is_valid_response(dev_req_t* resp, unsigned int request_id, 
		  unsigned int cmd)
{
	assert (resp);
	
	/* check that it's a response for this specific request */
	if (resp->id != request_id || resp->cmd != (cmd | REPLY) || 
	    resp->magic != DEV_FILE_MAGIC_COOKIE) {
		fprintf(stderr, "'%s': got invalid response '%d' for request '%d "
			"resp cmd is %x, resp magic is %x\n", __FUNCTION__, 
			resp->id, request_id, resp->cmd, resp->magic);
		return 0;
	}
	return 1; 
	
}


/* submit this request to the device file, and return the response in 'resp' */ 
/* return 1 on success, 0 on error */ 
int
dev_file_do_request(sct_ctrl_t ctrl, dev_req_t* req, dev_req_t* resp)
{
	int rc = 0; 
	int fd; 
	int retry_count = 1000; 
	/* open the device file */
	/* we only loop on error, to handle EBUSY */ 
	while ((fd = dev_file_open(get_device_file_name(ctrl))) == -1) {
	    /* if we get EBUSY, we should retry */ 
	    if (errno == EBUSY && --retry_count) {
		continue; 
	    } 
	    fprintf(stderr, "'%s': failed to open device file '%s' with error '%s'\n", 
		    __FUNCTION__, get_device_file_name(ctrl), strerror(errno));
	    return 0; 
	}
	if (!send_request(fd, req)){
		fprintf(stderr, "'%s': failed to send the request, with error '%s'\n", 
			__FUNCTION__, strerror(errno));
		goto close_dev; 
	}

	/* good, request sent, now let's wait for a response */
	memset(resp, 0, sizeof(*resp));
	if (!recv_response(fd, resp)){
		fprintf(stderr, "'%s': failed to receive a response from the kernel\n", 
			__FUNCTION__);
		goto close_dev; 
	}
	
	/* check that it's a valid response */ 
	if (!is_valid_response(resp, req->id, req->cmd))
		goto close_dev; 
	
	rc = 1; 

#if 0
	fprintf(stderr, "%s: about to close the fd\n", __FUNCTION__); 
#endif /* 0 */ 

 close_dev: 
	dev_file_close(fd); 
	
	return rc; 
}

/* this function should only be used for actions which don't require */ 
/* any data back except for an error code. For access to the response, */
/* use build_request/dev_file_do_request. */ 
/* we do most of the work here. we return 0 on success, */
/* -1 on genereal error and >0 on one of the protocol errors */
int 
dev_file_update_kernel(sct_ctrl_t ctrl, unsigned int cmd, 
		       int rule_id, sct_rule_when_t when, int syscall_id, 
		       unsigned char* buf, unsigned int length, 
		       int enable_rule)
{
	dev_req_t req, resp;
	int rc = -1;
	
	build_request(&req, cmd, rule_id, when, syscall_id, 
		      buf, length, enable_rule); 
	
	if (!dev_file_do_request(ctrl, &req, &resp))
		goto done;

	if (!is_valid_response(&resp, req.id, req.cmd))
		goto done; 
	
	rc = resp.result; 

	/* and now, cleanup */ 

	/* zero out the 'data' field which we dont own for the request */ 
	req.data = NULL; 

	/* and now do cleaup */ 
	destroy_request(&req); 
	destroy_request(&resp); 
 done:
    
	return rc;
}
