/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: parse_rulesbuf.c,v 1.3 2002/06/15 15:16:13 mulix Exp $
 */

#include "sct_rules/rule_engine_public.h"
#include "sct_rules/membuf.h"
#include "sct_rules_module/module_interface.h"

#include "sct_ctrl_lib.h"

/* general notes: */ 

/* we get a membuf which contains a list of serialized  */
/* rule lists, with the following structure: */

/* int magic cookie, int syscall id, int when, serialized rule list,  */
/* int magic cookie, int syscall id, int when, serialized rule list,  */
/* 0 */
 
/* and we return a list of syscall_rules_t structures, two lists per  */
/* syscall, for before and after rules. we return -1 on error, 0 if the */
/* list is empty and 1 on success  */

/* parse a single rule list out of the buffer */ 
/* return 1 on success, 0 if there's no list (marker == 0) and -1 on failure */ 
int parse_single_list(sct_membuf_t* mbuf, syscall_rules_t* p_list)
{
	sct_rule_when_t when; 
	int syscall_id; 
	int rc; 
	sct_rule_list_t* rules_list; 
	int magic; 

	if (!p_list || !mbuf) 
		return -1; 

	rc = sct_membuf_get_chunk(mbuf, &magic, sizeof(magic)); 
	if (rc != sizeof(magic))
		goto invalid_buffer; 

	if (magic != GET_RULES_MAGIC_COOKIE) {
		if (magic != 0) 
			goto invalid_buffer; 
		/* magic == 0 means no more data */ 
		return 0; 
	}	

	if (sct_membuf_get_chunk(mbuf, &syscall_id, 
				 sizeof(syscall_id)) != sizeof(syscall_id))
		goto invalid_buffer; 
	
	if (sct_membuf_get_chunk(mbuf, &when, sizeof(when)) != sizeof(when))
		goto invalid_buffer; 
	
	if (sct_deserialize_tracking_rule_list(mbuf, &rules_list) == -1)
		goto invalid_buffer; 

	/* we got it, now let's build the list */ 
	p_list->syscall_id = syscall_id; 
	p_list->when = when; 
	p_list->list = rules_list; 
	p_list->next = NULL; 

	return 1; 

 invalid_buffer:
	/* no cleanup necessary */ 
	return -1; 
}

void 
tear_down_list(syscall_rules_t* list)
{
	syscall_rules_t* head; 
	syscall_rules_t* t; 
	
	for (head = list; head; ) {
		sct_delete_tracking_rule_list(head->list); 
		t = head; 
		head = head->next; 
		/* don't delete the first one, it's user supplied */ 
		if (t != list)
			free(t); 
	}
	/* set the first one to bogus values */ 
	list->syscall_id = -1; 
	list->when = 0; 
	list->list = NULL; 
	list->next = NULL; 
}

int
parse_rulesbuf(sct_membuf_t* mbuf, syscall_rules_t* p_rules)
{
	int rc; 
	int count = 0; 
	syscall_rules_t* head = NULL; 
	syscall_rules_t* prev = NULL; 
	syscall_rules_t* slist; 

        
	if (!p_rules || !mbuf)
		return 0;

	do {
		/* create the next node in the list */ 
		if (!head) { /* first one, special handling */ 
			head = p_rules; 
			prev = NULL;
		} else {
			/* allocate a new structure for it */ 
			if (!(slist = malloc(sizeof(syscall_rules_t))))
				goto failed; 
			memset(slist, 0, sizeof(*slist)); 
			prev = head; 
			head = slist; 
		}
		
		/* try to parse the rule list out of the buffer */ 
		rc = parse_single_list(mbuf, head); 
		if (rc == -1) 			
			goto failed; /* uhoh, we failed. cleanup*/ 
		
		if (rc == 0)
			break; /* whee, we're done */ 

		/* add it to the list or rule lists, if it's not the first one */ 
		if (prev)
			prev->next = head; 

		++count; 
	} while (1); 

	/* success! return the number of rules in the list */
	return count; 

 failed: /* cleanup galore */ 
	tear_down_list(p_rules); 
	return -1; 
}
