/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: sct_ctrl_lib.c,v 1.17 2002/06/15 15:16:13 mulix Exp $
 */
#include "sct_ctrl_lib.h"
#include "sct_ctrl_lib_private.h"
#include "sysctl_impl.h"
#include "dev_file_impl.h"

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

/* init the library */
sct_ctrl_t
sct_ctrl_init(sct_ctrl_impl_t sct_impl)
{
    sct_ctrl_t ctrl = malloc(sizeof(*ctrl));
    if (!ctrl){
	return NULL;
    }

    /* better safe than sorry */
    memset(ctrl, 0, sizeof(struct sct_ctrl_type));

    /* default implementation type */
    switch (sct_impl) {
    case (SCT_CTRL_DEFAULT): /* fallthrough */
    case (SCT_CTRL_SYSCTL):
        fprintf(stderr, "sct_ctrl_lib is using sysctl for "
		"communications\n");
	ctrl->impl = SCT_CTRL_SYSCTL;
	ctrl->ops = &sysctl_ops; 
	break;
    case (SCT_CTRL_DEVICE_FILE):
        ctrl->impl = SCT_CTRL_DEVICE_FILE;
	ctrl->ops = &dev_file_ops; 
	fprintf(stderr, "sct_ctrl_lib is using the device file "
		"for communications\n");
	break;
    default:
	/* error, and a big one */
	fprintf(stderr, "sct_ctrl_init got unknown "
		"sct_ctrl_impl_t '%d'\n", sct_impl);
	free(ctrl);
	return NULL;
    }
    
    /* if our ops have an init function, call it */ 
    if (ctrl->ops && ctrl->ops->init)
	ctrl = ctrl->ops->init(sct_impl); 

    return ctrl;
}

/* shutdown the library */
int sct_ctrl_cleanup(sct_ctrl_t ctrl)
{
    /* if we have a cleanup function, call it */ 
    if (ctrl && ctrl->ops && ctrl->ops->cleanup)
	ctrl->ops->cleanup(ctrl); 
	
    if (ctrl)
	free(ctrl);

    return 1;
}

/* add a rule with the given rule id to the given syscall,    */
/* where the rule is made of this filter and this action      */
/* and 'enable rule' specifies if the rule should be          */
/* immediately enabled or not                                 */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_add_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when,
		  int syscall_id, 
		  sct_filter_node_t* filter, sct_action_t* act, 
		  int enable_rule)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->add_rule)
	return ctrl->ops->add_rule(ctrl, rule_id, when, 
				   syscall_id, filter, 
				   act, enable_rule);
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* delete from the given syscall the rule with this rule id   */
/* return 0 on failure, non 0 on success                      */
int
sct_ctrl_del_rule(sct_ctrl_t ctrl,
		  int rule_id,
		  sct_rule_when_t when,
		  int syscall_id)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->del_rule)
	return ctrl->ops->del_rule(ctrl, rule_id, when, syscall_id);
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* enable the rule with this id for this syscall              */
/* NOTE: currently NOT IMPLEMENTED                            */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_enable_rule(sct_ctrl_t ctrl,
		     int rule_id,
		     sct_rule_when_t when,
		     int syscall_id)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->enable_rule)
	return ctrl->ops->enable_rule(ctrl, rule_id, when, syscall_id);
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* disable the rule with this id for this syscall             */
/* NOTE: currently NOT IMPLEMENTED                            */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_disable_rule(sct_ctrl_t ctrl,
		      int rule_id,
		      sct_rule_when_t when,
		      int syscall_id)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->disable_rule)
	return ctrl->ops->disable_rule(ctrl, rule_id, when, syscall_id);
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* delete all of the rules for this system call.              */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->del_all_syscall_rules)
	return ctrl->ops->del_all_syscall_rules(ctrl, syscall_id);
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* delete all rules for ALL syscalls.                         */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_del_all_rules(sct_ctrl_t ctrl)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->del_all_rules)
	return ctrl->ops->del_all_rules(ctrl); 
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* print all of the current rules to the system's log file    */
/* return 0 on failure, non 0 on success                      */
int
sct_ctrl_print_all_rules(sct_ctrl_t ctrl)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->print_all_rules)
	return ctrl->ops->print_all_rules(ctrl); 
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}

/* returns the number of defined rules for all system calls */ 
int
sct_ctrl_get_all_rule_count(sct_ctrl_t ctrl)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->get_all_rule_count)
	return ctrl->ops->get_all_rule_count(ctrl); 
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}


/* returns a list of currently defined rules in p_rules (out  */ 
/* parameter). returns the number of rules in the list or -1  */
/* on error. (a list can have 0 rules)                        */
int
sct_ctrl_get_all_rules(sct_ctrl_t ctrl, syscall_rules_t* p_rules)
{
    if (!ctrl)
	return 0;

    if (ctrl->ops && ctrl->ops->get_all_rules)
	return ctrl->ops->get_all_rules(ctrl, p_rules); 
    
    fprintf(stderr, "%s: unknown control library implementation '%d'\n",
	    __FUNCTION__, ctrl->impl);
    return 0;
}
