/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU Lesser General Public License                       */

/* 
 * $Id: sysctl_impl.c,v 1.7 2002/06/15 15:16:13 mulix Exp $
 */

/* get the module's sysctl defines */
#include "sct_rules_module/module_interface.h"
#include "sct_rules/membuf.h"

#include "sct_ctrl_lib.h"
#include "sct_ctrl_lib_private.h"
#include "sysctl_impl.h"

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <linux/unistd.h>
#include <linux/sysctl.h>

_syscall1(int, _sysctl, struct __sysctl_args *, args);

sct_ctrl_ops_t sysctl_ops = {
	init: NULL, 
	cleanup: NULL, 
	add_rule: sysctl_add_rule, 
	del_rule: sysctl_del_rule,
	enable_rule: NULL, 
	disable_rule: NULL, 
	del_all_syscall_rules: sysctl_del_all_syscall_rules, 
	del_all_rules: sysctl_del_all_rules, 
	print_all_rules: sysctl_print_all_rules, 
	get_all_rule_count: sysctl_get_all_rule_count, 
	get_all_rules: sysctl_get_all_rules
}; 

/* returns 0 on error and 1 on success */
static int do_sysctl(sct_ctrl_t ctrl, unsigned char* buf, 
		     unsigned int buflen, int* sysctl_cmd, 
		     unsigned int sysctl_cmd_len)
{
    struct __sysctl_args args =
	{ sysctl_cmd, sysctl_cmd_len, NULL, 0, buf, buflen};
    
    int rc;
    
    assert (sysctl_cmd);
    
    if (!sysctl_cmd)
	return 0;
    
    rc = _sysctl(&args);
    if (rc != 0) 
	perror("sysctl");

    return (rc != -1) ? 1 : 0;
}

static int
sysctl_common(sct_ctrl_t ctrl, unsigned int sysctl_command, 
	      int rule_id, sct_rule_when_t when, int syscall_id, 
	      sct_filter_node_t* filter, sct_action_t* act,
	      int enable_rule)
{
    /* we ignore enable_rule for now */
    
    int serial_len;
    unsigned char buf[1000];
    int sysctl_cmd_struct[] = { CTL_TRACKER, sysctl_command };
    
    assert(rule_id >= 0 && syscall_id >= 0);
    
    if ((rule_id < 0) ||(syscall_id < 0))
	return 0;
    
    serial_len = 
	sct_serialize_tracking_rule(rule_id, when, syscall_id, act, 
				filter, buf, sizeof(buf));
    
    if (serial_len <= 0){
	fprintf(stderr, "failed to serialize tracking rule, "
		"error '%d'\n", serial_len);
	return 0;
    }

    return do_sysctl(ctrl, buf, serial_len, sysctl_cmd_struct, 
		     sizeof(sysctl_cmd_struct)/ sizeof(sysctl_cmd_struct[0]));
}

		     

/* add rule */
int 
sysctl_add_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when, int syscall_id, 
		sct_filter_node_t* filter, sct_action_t* act,
		int enable_rule)
{
    assert (filter && act); 
    if (!filter || !act)
	return 0;

    return sysctl_common(ctrl, CTL_TRACKER_CMD_ADD_RULE, 
			 rule_id, when, syscall_id, 
			 filter, act, enable_rule);
}

/* delete a rule */
int
sysctl_del_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when, int syscall_id)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_RULE, 
		       rule_id, when, syscall_id, 
		       /* filter node */NULL, 
		       /* action */NULL, 0);

}

/* enable a rule */
int 
sysctl_enable_rule(sct_ctrl_t ctrl,
		   int rule_id,
		   sct_rule_when_t when,
		   int syscall_id)
{
    fprintf(stderr, "eek! sct_ctrl_enable_rule is unimplented!\n");
    return 0;
}

/* disable a rule */
int 
sysctl_disable_rule(sct_ctrl_t ctrl,
		    int rule_id,
		    sct_rule_when_t when,
		    int syscall_id)
{
    fprintf(stderr, "eek! sct_ctrl_disable_rule is unimplented!\n");
    return 0;
}

/* delete all rules for this syscall */
int 
sysctl_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL, 
			 /* rule_id */ 0, /* when */ RW_BEFORE, syscall_id, 
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}

/* delete all rules for ALL syscalls */
int 
sysctl_del_all_rules(sct_ctrl_t ctrl)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES, 
			 /* rule_id */ 0, /* when */ RW_BEFORE,
			 /* syscall_id */ 0,
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}

/* print the current rules */
int
sysctl_print_all_rules(sct_ctrl_t ctrl)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_PRINT_ALL_RULES, 
			 /* rule_id */ 0, /* when */ RW_BEFORE,
			 /* syscall_id */ 0,
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}

/* get the count of all currently defined rules */ 
int
sysctl_get_all_rule_count(sct_ctrl_t ctrl)
{
    /* can't use sysctl_common, we need an out parameter */ 
    int sysctl_cmd_struct[] = { CTL_TRACKER, 
				CTL_TRACKER_CMD_GET_ALL_RULE_COUNT };

    int rc; 
    int sum; 
    size_t sz = sizeof(sum); 
    
    {
	struct __sysctl_args args =
	    { sysctl_cmd_struct, /* name */ 
	      sizeof(sysctl_cmd_struct)/ sizeof(sysctl_cmd_struct[0]), /* namelen */ 
	      &sum, /* output goes here */ 
	      &sz, /* overwritten with how big buf actually is */ 
	      NULL, /* newval */ 
	      0 /* newlen */ 
	    };
	
	rc = _sysctl(&args); 
	if (rc) { /* woops, sysctl failed. try to figure out why */ 
		perror("sysctl");
		return -1;
	}
    }
    return sum; 
}

/* get all of the current rules */
int
sysctl_get_all_rules(sct_ctrl_t ctrl, syscall_rules_t* p_rules)
{
    /* can't use sysctl_common, we need an out parameter */ 
    int sysctl_cmd_struct[] = { CTL_TRACKER, 
				CTL_TRACKER_CMD_GET_ALL_RULES };

    /* 16 k should be enough until we implement retry on buf too small */ 
    unsigned char kernbuf[16 * 1024]; 
    int sbuf = sizeof(kernbuf); 
    
    sct_membuf_t* mbuf; 
    int rc; 
    
    {
	struct __sysctl_args args =
	    { sysctl_cmd_struct, /* name */ 
	      sizeof(sysctl_cmd_struct)/ sizeof(sysctl_cmd_struct[0]), /* namelen */ 
	      kernbuf, /* output goes here */ 
	      &sbuf, /* overwritten with how big buf actually is */ 
	      NULL, /* newval */ 
	      0 /* newlen */ 
	    };
	
	rc = _sysctl(&args); 
	if (rc) { /* woops, sysctl failed. try to figure out why */ 
            if (errno == ENOSPC) {
		    /* TODO: retry with a bigger buffer */ 
		    printf("syscall_get_all_rules: buffer size %d too small. \n", sbuf); 
            } else {
		    perror("sysctl");
	    }
	    return -1;
	}
    }

    /* put the buffer we got back into a membuf */ 
    mbuf = sct_membuf_create_with_data(kernbuf, sbuf, NULL, NULL); 
    if (!mbuf)
	    return -1; 

    rc = parse_rulesbuf(mbuf, p_rules); 
    if (rc == -1) {
	    fprintf(stderr, "failed to parse rules buffer\n"); 
	    return -1; 
    } else {
	    fprintf(stderr, "parse_rulesbuf returned %d\n", rc); 
    }
	    
    sct_membuf_destroy(mbuf); 

    return rc; 
}
