/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/* 
 * $Id: tester.c,v 1.29 2002/06/15 15:16:14 mulix Exp $
 */ 

/* 
 * a testing program for the kernel module. to add more tests, write a
 * test function and add a call to it in main. your funtion should return
 * 0 if the test was succesful and anything else otherwise. 
 */
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <assert.h>
#include <signal.h>
#include <setjmp.h>

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/rule_engine_public.h"

#include "tester.h"

const char* progname;
sct_ctrl_t ctrl;
static char* dirname; 
static char* topdir; 
static int did_cleanup; 

static void dir_setup(void), dir_cleanup(void); 
 static void comm_setup(sct_ctrl_impl_t impl); 
static void comm_cleanup(sct_ctrl_t ctrl); 

/* we need to protect from SIGINT, SIGQUIT and SIGTERM */ 
struct sigsave {
	int signum; 
	struct sigaction sigact; 
} oldsigs[3]; 

sigjmp_buf env; 
static void sighandler(int signum)
{
	fprintf(stderr, "sighandler: caught signal %d\n", signum); 
	
	siglongjmp(env, 1);
}

void signals_setup(void)
{
        /* we need to protect from SIGINT, SIGQUIT and SIGTERM */ 
	struct sigaction act; 
	
	memset(&act, 0, sizeof(act)); 
	memset(&oldsigs, 0, sizeof(oldsigs)); 

	act.sa_handler = &sighandler; 
	sigemptyset(&act.sa_mask); 

	if (sigaction(SIGINT, &act, &oldsigs[0].sigact)) {
		perror("sigaction"); 
		return; 
	}
	oldsigs[0].signum = SIGINT; 

	if (sigaction(SIGQUIT, &act, &oldsigs[1].sigact)) {
		perror("sigaction"); 
		return; 
	}
	oldsigs[1].signum = SIGQUIT; 
	
	if (sigaction(SIGTERM, &act, &oldsigs[2].sigact)) {
		perror("sigaction"); 
		return; 
	}
	oldsigs[2].signum = SIGTERM; 
}

void signals_cleanup(void)
{
	int i; 
	for (i = 0; i < sizeof(oldsigs)/sizeof(oldsigs[0]); ++i) {
		if (oldsigs[i].signum) {
			if (sigaction(oldsigs[i].signum, 
				      &oldsigs[0].sigact, 
				      NULL)) {
				perror("sigaction"); 
			}
		}
	}
}

void 
_die(const char* func, const char* file,
     unsigned int line, const char* fmt, ...)
{
    va_list ap; 

    fprintf(stdout, "[%s:%d:%s] ",file, line, func);
    
    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);

    assert(0);
}

void 
_logerr(const char* func, const char* file,
	unsigned int line, const char* fmt, ...)
{
    va_list ap; 
    
    printf("[%s:%d:%s] ",file, line, func);

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
}

static void 
comm_setup(sct_ctrl_impl_t impl)
{
    /* setup module */
    ctrl = sct_ctrl_init(impl);
    if (!ctrl){
	die("failed to init sct_ctrl_lib\n");
    }
    
    /* del module rules */
    if (!sct_ctrl_del_all_rules(ctrl)){
	die("failed to delete all rules\n");
    }
}

static void comm_cleanup(sct_ctrl_t ctrl)
{
	sct_ctrl_cleanup(ctrl); 
}

static void dir_setup(void)
{
	struct stat buf; 
	char tmpl[32];

	topdir = getcwd(NULL, 0); 
	if (!topdir) 
		die("getcwd failed"); 

	snprintf(tmpl, sizeof(tmpl), "%sXXXXXX", "testdir");
	
	if (!(dirname = strdup(mkdtemp(tmpl))))
		die("mkdtemp: %s", strerror(errno));
	if (stat(dirname, &buf))
		die("stat: %s", strerror(errno)); 
	if ((!S_ISDIR(buf.st_mode)) || (buf.st_mode & (S_IRWXG | S_IRWXO)))
		die("%s is not a dir or wrong permissions"); 

	if (chdir(dirname))
		die("chdir: %s", strerror(errno)); 
}

static void dir_cleanup(void)
{
	DIR* d; 
	struct dirent* e; 

	if (did_cleanup)
		return; 

	if (!(d = opendir("./"))) {
		perror("opendir"); 
		return; 
	}
	
	/* we ignore errors */ 
	while ((e = readdir(d))) {
		if (!strcmp(e->d_name, ".") || !strcmp(e->d_name, ".."))
			continue; 

		if (remove(e->d_name))
			perror("remove"); 

		/* we removed an entry, so rewind the dir stream */ 
		rewinddir(d); 
	}

	closedir(d); 

	/* back to our starting directory */ 
	chdir(topdir); 

	if (rmdir(dirname))
		perror("rmdir"); 

	free(dirname); 
	dirname = NULL; 
	free(topdir); 
	topdir = NULL; 
}

int 
main(int argc, char* argv[])
{
    sct_ctrl_impl_t impl = SCT_CTRL_DEFAULT; 

    progname = argv[0];
    
    if (argc >= 3 && !strcmp(argv[1], "-c")) {
	if (!strcmp(argv[2],"device_file"))
	    impl = SCT_CTRL_DEVICE_FILE; 
	else if (!strcmp(argv[2],"sysctl"))
            impl = SCT_CTRL_SYSCTL; 
	else {
	    fprintf(stdout, "usage: %s [-c device_file|sysctl]\n",
		    progname); 
	    exit(1); 
	}
    }

    signals_setup(); 

    if (sigsetjmp(env, 1))
	    goto shutdown;

    comm_setup(impl);

    dir_setup(); 
   
    /* sct_rules tests */
    fprintf(stdout, "starting sct_rules tests\n");

    TEST(test_filter_eval_bit());
    
    TEST(test_eval_complex_filter());
    
    TEST(test_filter_serialize_deserialize());
    
    TEST(test_rules_serialize_deserialize());
    
    TEST(test_tracking_rule_add_delete());

    TEST(test_tracking_rule_add_delete_empty_list());

    TEST(test_kernel_variables());
    
    TEST(test_short_circuiting());

    TEST(test_membuf()); 

    TEST(test_tracking_rule_list()); 

    /* get_rules tests */ 

    fprintf(stdout, "starting get_rules tests\n");

    TEST(test_get_rules()); 

    /* buffer support tests */ 

    fprintf(stdout, "starting buffer support tests\n"); 

    TEST(test_void_ptr_buffer()); 

    TEST(test_buffer_match()); 

    /* module tests */ 

    fprintf(stdout, "starting module tests\n");

    TEST(test_simple_syscall_usage());

    TEST(test_rule_register_deregister());

    TEST(test_all_syscalls_usage());
 
#ifdef TEST_ME_HARDER /* disable this test by default, since it can be destructive */
     TEST(test_settimeofday_usage()); 
#endif  /* TEST_ME_HARDER */ 

    fprintf(stdout, "tests ended succesfully!\n");

 shutdown:
    
    dir_cleanup(); 

    comm_cleanup(ctrl); 

    signals_cleanup(); 

    return 0;
}

